from torch import nn
from .modules import MishLayer


class Classifier(nn.Module):
	def __init__(self, inp_size: int, out_size: int, hid_size: int = 512):
		super(Classifier, self).__init__()

		self.model = nn.Sequential(
			#BatchSwapNoise(0.15),
			nn.Linear(inp_size, hid_size),
			nn.BatchNorm1d(hid_size),
                        MishLayer(),
			nn.Linear(hid_size, hid_size//2),
                        MishLayer(),
			nn.BatchNorm1d(hid_size//2),
			nn.Linear(hid_size//2, hid_size//4),
                        MishLayer(),
                        nn.BatchNorm1d(hid_size//4),
                        nn.Linear(hid_size//4, out_size)
			#nn.Softmax(-1),
		)

	def forward(self, z):
		validity = self.model(z)
		return validity



from itertools import combinations
from ignite.engine import _prepare_batch
from torch import mean, exp, unique, cat, isnan
from torch import device as torch_device
from torch import norm as torch_norm
from torch.nn import NLLLoss, Linear, CrossEntropyLoss, MSELoss
from torch.nn.utils import clip_grad_norm_
from torch.optim.lr_scheduler import ExponentialLR, CosineAnnealingLR
from torch.optim import Adam

from .model import RAdam
import torch_optimizer as toptim
from .model.losses import VariationalLoss, mmd_criterion
from .data.utils import create_random_classes, add_noise




class VAEUpdater:
    '''
    Class contains one iteration of learning loop.
    '''
    def __init__(self, vae_model, disc, cfg, device=None, cuda=False):
        self.cfg = cfg
        self.noise_beta = self.cfg.noise_beta

        self.model = vae_model
        self.model_linear_params = cat([x.view(-1) for x in self.model.parameters()])#for l1/l2 reg
        self.latent_discrim = disc

        self.optimizer = toptim.Ranger(self.model.parameters(), lr=cfg.vae_lr, weight_decay=cfg.vae_decay)
        self.optimizer_discrim = toptim.Ranger(self.latent_discrim.parameters(), lr=cfg.disc_lr, weight_decay=cfg.disc_decay)
        self.model_scheduler =  CosineAnnealingLR(self.optimizer, 5) #ExponentialLR(self.optimizer, gamma=cfg.model_scheduler_gamma)
        self.latent_discrim_scheduler = ExponentialLR(self.optimizer_discrim, gamma=cfg.discr_scheduler_gamma)

        self.loss_function = VariationalLoss(self.cfg.vae_beta, 1.)
        self.discrim_loss = NLLLoss()

        self.form_consistency_loss = CrossEntropyLoss()

        self.device = device
        self.cuda = cuda

    def __call__(self, engine, batch):
        self.noise_beta = self.noise_beta * self.cfg.decay_beta

        data, ohe = _prepare_batch(batch, device=self.device)
        #Autoencoder training
        if engine.iteration % 2 == 0:
            self.model.train()
            self.latent_discrim.eval()

            self.optimizer.zero_grad()

            #encoding expression batch with initial classes and evaluating the reconstruction loss
            a2b_mu, a2b_logvar, a2b_form = self.model.encode(data, ohe)
            a2b_latents = self.model.reparameterize(a2b_mu, a2b_logvar)
            a2a_expression, mmd_values = self.model.decode(a2b_latents, ohe)
            reconstruction_loss_on_batch = self.loss_function(a2a_expression, data, a2b_mu, a2b_logvar)

            #mmd loss
            av_mmd_loss = 0.
            batch_indices = ohe.argmax(1)
            unique_classes = unique(batch_indices)
            count_comb = 0
            for feature1, feature2 in combinations(unique_classes, 2):
                count_comb += 1
                x1 = mmd_values[batch_indices == feature1]
                x2 = mmd_values[batch_indices == feature2]
                min_len = min(x1.shape[0], x2.shape[0])
                av_mmd_loss += mmd_criterion(x1[:min_len], x2[:min_len], self.cuda, mu=self.cfg.kernel_mu, sigma=1e-6).mean()
            av_mmd_loss /= count_comb

            #first style transfer
            transfer_classes = create_random_classes(data.size(0), ohe.shape[1])
            if self.cuda:
                transfer_classes = transfer_classes.cuda()
            a2b_expression = self.model.decode(a2b_latents, transfer_classes)[0]
            #second style transfer
            b2a_mu, b2a_logvar, b2a_form = self.model.encode(a2b_expression, ohe)
            b2a_latents = self.model.reparameterize(b2a_mu, b2a_logvar)
            b2a_expression = self.model.decode(b2a_latents, ohe)[0]
            cyclic_loss_on_batch = self.loss_function.reconstruction_loss(b2a_expression, data)

            #form consistency loss
            #form_consistency_loss_on_batch = self.form_consistency_loss(b2a_form, ohe.argmax(1))
            #form_consistency_loss_on_batch = self.form_consistency_loss(b2a_form, transfer_classes.argmax(1))
            #form_consistency_loss_on_batch = self.form_consistency_loss(a2b_form, ohe.argmax(1))
            a2a_mu, a2a_logvar, a2a_form = self.model.encode(a2a_expression, ohe)
            a2a_latents = self.model.reparameterize(a2a_mu, a2a_logvar)

            form_consistency_loss_on_batch = self.form_consistency_loss(a2b_form, ohe.argmax(1)) + self.form_consistency_loss(b2a_form, transfer_classes.argmax(1)) + MSELoss()(a2a_latents, b2a_latents)

            #adversarial part
            mu, logvar, form_ = self.model.encode(data, ohe)
            latents = self.model.reparameterize(mu, logvar)
            latents = self.model.reparameterize(mu, logvar)
            #next_layers_out = self.model.decode(latents, ohe)[1]

            #adv_loss_g = discrim_loss(latent_discrim(latents), ohe.argmax(1))
            #loss-=20.*adv_loss_g

            noisy_latents = add_noise(latents, self.noise_beta, self.cuda)
            #noisy_latents = add_noise(next_layers_out)
            discrim_preds = self.latent_discrim(noisy_latents)
            #shannon entropy
            adv_loss_g = mean(discrim_preds * exp(discrim_preds))

            #weight regularization
            l1_regularization = torch_norm(self.model_linear_params, 1)

            loss = reconstruction_loss_on_batch + self.cfg.cyclic_weight*cyclic_loss_on_batch + \
                    self.cfg.adv_weight*adv_loss_g - self.cfg.mmd_weight*av_mmd_loss + \
                    self.cfg.l1_weight*l1_regularization +self.cfg.form_consistency_weight*form_consistency_loss_on_batch
            #print('Cyclic:', cyclic_loss_on_batch.item())
            #print('Sum:', loss.item())
            loss.backward()
            engine.state.adversarial_loss = loss.item()
            #Clip gradient
            clip_grad_norm_(self.model.parameters(), self.cfg.clip_value)
            self.optimizer.step()

        elif engine.iteration % 2 == 1:
            #discriminator learning part
            self.model.eval()
            self.latent_discrim.train()

            self.optimizer_discrim.zero_grad()
            mu, logvar, form_ = self.model.encode(data, ohe)
            latents = self.model.reparameterize(mu, logvar)
            noisy_latents = add_noise(latents, self.noise_beta)
            adv_loss_d = self.discrim_loss(self.latent_discrim(noisy_latents.detach()), ohe.argmax(1))

            adv_loss_d.backward()
            clip_grad_norm_(self.latent_discrim.parameters(), self.cfg.clip_value)
            self.optimizer_discrim.step()
#            self.latent_discrim_scheduler.step(adv_loss_d.item())
#        if engine.iteration > 4500:
#            self.model_scheduler.step()
#        self.model_scheduler.step()
#        self.latent_discrim_scheduler.step()
        engine.iteration += 1


class Validator:
    def __init__(self, model, discriminator, vae_beta, cuda=False, device=None):
        self.loss_function = VariationalLoss(vae_beta)
        self.model = model
        self.latent_discrim = discriminator
        self.cuda = cuda
        self.device = device
        self.discrim_loss = NLLLoss()


    def __call__(self, engine, batch):
        self.model.eval()
        if self.latent_discrim:
            self.latent_discrim.eval()

        val_expression_tensor, val_class_ohe_tensor = _prepare_batch(
            batch,
            device=self.device
        )

        recon_val, mu_val, logvar_val = self.model(val_expression_tensor,
                                                   val_class_ohe_tensor)
        mmd_values = None
        if isinstance(recon_val, tuple):
            mmd_values = recon_val[1]
            recon_val = recon_val[0]#only reconstr values, without mmd values
        latents_val = self.model.reparameterize(mu_val, logvar_val)
        reconstruction_loss_val = self.loss_function.reconstruction_loss(recon_val, val_expression_tensor)

        av_mmd_loss = 0.
        batch_indices = val_class_ohe_tensor.argmax(1)
        unique_classes = unique(batch_indices)
        count_comb = 0
        for feature1, feature2 in combinations(unique_classes, 2):
            x1 = mmd_values[batch_indices == feature1]
            x2 = mmd_values[batch_indices == feature2]
            min_len = min(x1.shape[0], x2.shape[0])
            av_mmd_loss += mmd_criterion(x1[:min_len], x2[:min_len], self.cuda).mean()
            count_comb += 1
        av_mmd_loss /= count_comb

        engine.state.losses = {}
        if self.latent_discrim:
            discrim_preds_val = self.latent_discrim(latents_val)
            discrim_loss_val = self.discrim_loss(discrim_preds_val.detach(), val_class_ohe_tensor.argmax(1))
            discrim_val_acc = (val_class_ohe_tensor.argmax(1) == discrim_preds_val.argmax(1)).float().mean()

            #first style transfer
            transfer_classes_val = create_random_classes(val_expression_tensor.shape[0],
                                                                    val_class_ohe_tensor.shape[1])
            if self.cuda:
                transfer_classes_val = transfer_classes_val.cuda()
            a2b_expression = self.model.decode(latents_val, transfer_classes_val)[0]
            #second style transfer
            b2a_mu, b2a_logvar, form_ = self.model.encode(a2b_expression, transfer_classes_val)
            b2a_latents = self.model.reparameterize(b2a_mu, b2a_logvar)
            b2a_expression = self.model.decode(b2a_latents, val_class_ohe_tensor)[0]
            cyclic_loss_val = self.loss_function.reconstruction_loss(b2a_expression, val_expression_tensor)
            engine.state.losses['discrim loss'] = discrim_loss_val.item()
            engine.state.losses['cyclic'] = cyclic_loss_val.item()
            engine.state.losses['discrim accuracy'] = discrim_val_acc if isinstance(discrim_val_acc, float) else discrim_val_acc.item()

        engine.state.losses['reconstruction'] = reconstruction_loss_val.item()
        engine.state.losses['mmd'] = av_mmd_loss.item() if not isinstance(av_mmd_loss, float) else av_mmd_loss


class trVAEUpdater:
    def __init__(self, model, cfg, device=None, cuda=False):
        self.model = model
        self.cfg = cfg
        self.cuda = cuda
        self.optimizer = Adam(self.model.parameters(), lr = self.cfg.lr)
        self.device = device
        self.vae_loss = VariationalLoss(cfg.vae_beta, cfg.rec_coef)

    def __call__(self, engine, batch):
        expression, ohe = _prepare_batch(batch, device=self.device)
        batch_indices = ohe.argmax(1).float().view(-1, 1)

        mu, logvar = self.model.encode(expression, batch_indices)
        latents = self.model.reparameterize(mu, logvar)
        reconstructed, mmd_values = self.model.decode(latents, batch_indices)

        vae_loss = self.vae_loss(reconstructed, expression, mu, logvar)

        av_mmd_loss = 0.
        unique_classes = unique(batch_indices)
        count_comb = 0

        for feature1, feature2 in combinations(unique_classes, 2):
            count_comb += 1
            x1 = mmd_values[(batch_indices == feature1).reshape((-1,))]
            x2 = mmd_values[(batch_indices == feature2).reshape((-1,))]
            min_len = min(x1.shape[0], x2.shape[0])
            av_mmd_loss += mmd_criterion(x1[:min_len], x2[:min_len], self.cuda,
                                         mu=self.cfg.kernel_mu, sigma=1e-6).mean()
        av_mmd_loss /= count_comb

        loss_on_batch = vae_loss + self.cfg.mmd_weight * av_mmd_loss

        self.optimizer.zero_grad()
        loss_on_batch.backward()
        self.optimizer.step()




import argparse
import importlib
from pathlib import Path

from scvi.dataset import (
    PreFrontalCortexStarmapDataset,
    RetinaDataset,
    PbmcDataset,
    CsvDataset
)

from datasets import (
    BermudaDataset
)
from utils import (
    predefined_preprocessing
)
from config import Config
from experiment import Experiment

from benchmarkst import (
    benchmark_scvi,
    benchmark_scgen,
    benchmark_trvae,
    benchmark_stvae
)


DIRPATH = Path('./').absolute()

BENCHMARK_FUNCTIONS = (
    benchmark_stvae,
    benchmark_scvi,
    benchmark_scgen,
    benchmark_trvae
)
FRAMEWORKS = (
    'stvae',
    'scvi',
    'scgen',
    'trvae'
)
#EPOCHS = (600, 100, 100, 300)
EPOCHS = (1, 1, 1, 1)

datasets = {
    'scvi_pbmc': PbmcDataset(),
    'bermuda_pbmc': CsvDataset(
        str(DIRPATH / './pbmc/expression.csv'),
        labels_file = str(DIRPATH / './pbmc/labels.csv'),
        batch_ids_file = str(DIRPATH / './pbmc/batches.csv'),
        gene_by_cell = False
    ),
    'mouse': CsvDataset(
        str(DIRPATH / './mouse_genes/ST1 - original_expression.csv'),
        labels_file = str(DIRPATH / './mouse_genes/labels.csv'),
        batch_ids_file = str(DIRPATH / './mouse_genes/batches.csv'),
        gene_by_cell = False
    ),
    #'pancreas': BermudaDataset('./pancreas/muraro_seurat.csv'),
    'retina': RetinaDataset(),
    'starmap': PreFrontalCortexStarmapDataset(),
}

parser = argparse.ArgumentParser(description="A way to define variables for \
                                 training, tests and models")
parser.add_argument('--metrics_dir', type=str, help="Path to save metrics file \
                    \nDisabled if save_metrics flag is False.\n\nDefault value \
                    is './'")
parser.add_argument('--save_metrics', type=bool, help='Boolean flag determines \
                    whether metrics should be saved.\n\n Default value is True')
parser.add_argument('--custom_config', type=str, help='The path to the \
                    configuration file to be used. If equal to None, then \
                    the config located in the same folder will be used.',
                    default=None)
args = parser.parse_args()

cfg = Config()
with Experiment('', cfg) as exp:
    if args.metrics_dir:
        cfg.metrics_dir = args.metrics_dir
    if args.custom_config:
        try:
            spec = importlib.util.spec_from_file_location(args.custom_config)
            config_module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(config_module)
            cfg = config_module.Config()
        except:
            raise ValueError(f"Cannot import config module from {args.custom_config}")

    for log_name, dataset in datasets.items():
        for bench_func, epoch, framework in zip(BENCHMARK_FUNCTIONS,
                                                EPOCHS,
                                                FRAMEWORKS):
            cfg.epochs = epoch
            data = predefined_preprocessing(dataset, framework=framework)
            #data = dataset
            bench_func(data, log_name, cfg)



from data import make_anndata
from utils import load_datasets, scvi_anndata, get_high_variance_genes
from train_script import train
from test_script import test
from config import Config
from model.trvae import trVAE
from experiment import Experiment

from scvi.dataset import (
    PreFrontalCortexStarmapDataset,
    FrontalCortexDropseqDataset,
    RetinaDataset,
    HematoDataset,
    AnnDatasetFromAnnData,
)
import scgen
from model import SCGene
import os
import json
import numpy as np
from torch import zeros, Tensor, LongTensor, log, cuda, save, load
import scanpy as sc
import trvaep
from functools import partial
from pathlib import Path


def benchmark_scgen(dataset, log_name, cfg, **kwargs):
    ds = dataset
    n_genes = min(ds.X.shape[1], cfg.n_genes)

    scvai_genes, scvai_batches_ind, scvai_labels_ind = get_high_variance_genes(
        ds.X, ds.batch_indices, ds.labels, n_genes=n_genes, argmax=False
    )

    cfg.count_classes = np.unique(ds.batch_indices).shape[0]
    cfg.count_labels = np.unique(ds.labels).shape[0]
    cfg.input_dim = int(scvai_genes.shape[1])

    data = load_datasets(
        cfg, True, True, (scvai_genes, scvai_batches_ind, scvai_labels_ind)
    )
    dataloader_train = data[0]
    dataloader_val = data[1]
    dataloader_test = data[2]
    annot_train = data[3]
    annot_test = data[4]

    # train data
    x, batch_ind, celltype = annot_train.dataset.tensors
    batch_ind = batch_ind.argmax(dim=1)
    celltype = celltype.argmax(dim=1)
    anndata_train = make_anndata(
        x.cpu().numpy(),
        batch_ind.cpu().numpy(),
        "condition",
        celltype.cpu().numpy(),
        "cell_type",
    )
    sc.pp.normalize_per_cell(anndata_train)
    sc.pp.log1p(anndata_train)
    """
    # validation data
    x_valid, batch_valid, celltype_valid = dataloader_val.dataset.tensors
    batch_valid = batch_valid.argmax(1)
    celltype_valid = celltype_valid.argmax(1)
    anndata_valid = make_anndata(x_valid.cpu().numpy(),
                                 batch_valid.cpu().numpy(), 'condition',
                                 celltype_valid.cpu().numpy(), 'cell_type')
    sc.pp.normalize_per_cell(anndata_valid)
    sc.pp.log1p(anndata_valid)
    """
    # test data
    x_test, batch_ind_test, celltype_test = annot_test.dataset.tensors
    batch_ind_test = batch_ind_test.argmax(dim=1)
    celltype_test = celltype_test.argmax(dim=1)
    anndata_test = make_anndata(
        x_test.cpu().numpy(),
        batch_ind_test.cpu().numpy(),
        "condition",
        celltype_test.cpu().numpy(),
        "cell_type",
    )
    sc.pp.normalize_per_cell(anndata_test)
    sc.pp.log1p(anndata_test)

    n_conditions = anndata_train.obs["condition"].unique().shape[0]
    x_test = anndata_test.X
    batch_ind_test_tmp = anndata_test.obs["condition"]
    batch_ind_test = zeros(batch_ind_test_tmp.shape[0], cfg.count_classes)
    batch_ind_test = batch_ind_test.scatter(
        1, LongTensor(batch_ind_test_tmp.astype("uint16")).view(-1, 1), 1
    ).numpy()
    celltype_test_tmp = anndata_test.obs["cell_type"]
    celltype_test = zeros(celltype_test_tmp.shape[0], cfg.count_labels)
    celltype_test = celltype_test.scatter(
        1, LongTensor(celltype_test_tmp.astype("uint16")).view(-1, 1), 1
    ).numpy()

    # model = scgen.VAEArith(x_dimension=cfg.input_dim)
    model = SCGene(
        cfg.input_dim,
        latent_dim=cfg.bottleneck,
        device="cuda"
    )

    print("Training...")
    model.train(train_data=anndata_train, n_epochs=cfg.epochs)

    print("Tests...")

    print("Dataset:", log_name)
    res = test(
        cfg,
        model,
        None,
        annot_train,
        x_test,
        batch_ind_test,
        celltype_test
        # styletransfer_test_expr,
        # styletransfer_test_class,
        # styletransfer_test_celltype
    )

    metrics_path = Path(cfg.metrics_dir) / 'scGen'
    metrics_path.mkdir(parents=True, exist_ok=True)
    with open(metrics_path / (log_name + ".json"), "w") as file:
        json.dump(res, file, indent=4)

    del ds  # , res
    del model
    # del styletransfer_test_expr
    # del styletransfer_test_class
    # del styletransfer_test_celltype
    del data
    del dataloader_train, dataloader_val, dataloader_test
    del annot_train, annot_test
    del scvai_genes, scvai_batches_ind, scvai_labels_ind
    cuda.empty_cache()



from typing import Optional, Sized, Union
import numpy as np
import scanpy as sc
from sklearn.preprocessing import OneHotEncoder
from scvi.dataset import GeneExpressionDataset

from .datasets import ScratchDataset
from .utils import make_anndata, scvi_anndata


def predefined_preprocessing(
    data: Sized, framework: str, data_format: Optional[str] = "adaptive"
):
    """Prepare for any supported frameworks
    :Param data: data of supoorted type [scanpy.Anndata,
        tuple(at least one numpy.ndarray), scVI dataset like class]
    :Param framework: flag defining one of supported frameworks
    :Param data_format: define what kind of data is passed
    Note: Unrecognized framework param occurs NotImplementedError
        Unrecognized data_format param occurs ValueError
        If type of data unrecognized it occurs TypeError
    """

    _supported_formats = ["adaptive", "raw", "scvi", "anndata"]
    _supported_frameworks = ["stvae", "scvi", "scgen", "trvae"]
    framework = framework.lower()

    if not framework in _supported_frameworks:
        raise NotImplementedError(
            f"{framework} isn't supported value \
                                    {' '.join(_supported_frameworks)}"
        )
    if not data_format in _supported_formats:
        raise ValueError(
            f"{data_format} is not supported. \
                            Supported values are {''.join(_supported_formats)}"
        )

    def scvi_format(data, framework):
        if framework == "scvi":
            return data
        elif framework == "stvae":
            expression = np.log(data.X + 1.0)
            return ScratchDataset(expression, data.batch_indices, data.labels)
        data = scvi_anndata(data)  # trvae and scgen
        return data

    def raw_format(data, framework):
        if framework == "scvi":
            pass  # find a way to make scvi dataset from scratch by native classes
        elif framework == "stvae":
            if len(data) != 3:
                raise ValueError(f"Exprected data len = 3, got {len(data)}")
            expression = np.log(data[0] + 1.0)
            return ScratchDataset(
                expression, data[1], data[2]  # expression, batch indices, label indices
            )
        elif framework in ("scgen", "trvae"):
            if len(data) < 1:
                raise ValueError("Data must not be empty")
            _keys = ("batch_index", "cell_info", "variables_info")
            anndata_arguments = {_keys[i]: data[i + 1] for i in range(len(data))}
            if len(data) > 1:
                anndata_arguments["batch_name"] = "condition"
            if len(data) > 2:
                anndata_arguments["cell_info_name"] = "cell_type"

            return make_anndata(data[0], **anndata_arguments)

    if data_format == "scvi":
        data = scvi_format(data, framework)
    elif data_format == "raw":
        data = raw_format(data, framework)
    # do anndata
    elif data_format == "adaptive":
        # Attention: GeneExpressionDataset is Sized
        if isinstance(data, GeneExpressionDataset):
            data = scvi_format(data, framework)
        elif isinstance(data, Iterable):
            data = raw_format(data, framework)
        else:
            raise TypeError(f"Unrecognized data type: {type(data)}")
        # do AnnData
    if framework in ("scgen", "trvae"):
        sc.pp.normalize_per_cell(data)
        sc.pp.log1p(data)

    return data

def get_high_variance_genes(
    dataset,
    genes: Optional[Union[list, tuple, np.ndarray]] = None,
    n_genes: Optional[int] = None,
    min_disp: Optional[float] = None,
    max_disp: Optional[float] = None,
    argmax: bool = True,
) -> tuple:
    assert n_genes or min_disp or max_disp, "Choose smth"

    expression = dataset.X
    class_ohe = dataset.batch_indices
    cell_type = dataset.labels
    genes_data = None
    if not genes is None:
        genes_data = [("genes", genes)]
        if not isinstance(genes, np.ndarray):
            genes = np.array(genes)
    if argmax:
        class_ohe = class_ohe.argmax(1)
        cell_type = cell_type.argmax(1)
    anndata = make_anndata(
        expression, class_ohe, "condition", cell_type, "cell_type", genes_data
    )
    sc.pp.filter_genes_dispersion(
        anndata, n_top_genes=n_genes, min_disp=min_disp, max_disp=max_disp
    )

    expr = anndata.X
    expr = expr if isinstance(expr, np.ndarray) else np.array(expr)
    ohe = np.array(anndata.obs["condition"].tolist())
    ohe = OneHotEncoder(sparse=False).fit_transform(ohe.reshape(-1, 1))
    celltype = np.array(anndata.obs["cell_type"])
    celltype = OneHotEncoder(sparse=False).fit_transform(celltype.reshape(-1, 1))
    if not genes is None:
        return expr, ohe, celltype, anndata.var["genes"]

    return ScratchDataset(expr, ohe, celltype)


from utils import load_datasets, scvi_anndata, get_high_variance_genes
from model.ml import entropy_batch_mixing
from model.ml import get_knn_purity
from train_script import train
from test_script import test, train_classifiers
from config import Config
from experiment import Experiment

from scvi.dataset import (
    PreFrontalCortexStarmapDataset,
    FrontalCortexDropseqDataset,
    RetinaDataset,
    HematoDataset,
    CsvDataset,
    AnnDatasetFromAnnData,
)
from scvi.models import SCANVI, VAE
from scvi.inference import (
    UnsupervisedTrainer,
    JointSemiSupervisedTrainer,
    SemiSupervisedTrainer,
)
from sklearn.model_selection import train_test_split
import os
import json
import numpy as np
import torch
from torch import zeros, Tensor, LongTensor, log, cuda, save, load
from tsnecuda import TSNE
from matplotlib import pyplot as plt
from tqdm import tqdm
from pathlib import Path
import scanpy as sc


def benchmark_scvi(dataset, dataset_name, cfg, **kwargs):
    log_name = dataset_name
    n_genes = min(dataset.X.shape[1], cfg.n_genes)

    vae = VAE(
        dataset.nb_genes,
        n_batch=dataset.n_batches,
        n_labels=dataset.n_labels,
        n_hidden=128,
        n_latent=30,
        n_layers=2,
        dispersion="gene",
    )

    trainer = UnsupervisedTrainer(vae, dataset, train_size=0.75)
    n_epochs = cfg.epochs if not "epochs" in kwargs else kwargs["epochs"]
    trainer.train(n_epochs=n_epochs)

    full = trainer.create_posterior(
        trainer.model, dataset, indices=np.arange(len(dataset))
    )
    latents, batch_indices, labels = full.sequential().get_latent()

    res = {}
    res["knn purity"] = []
    res["entropy batch mixing"] = []
    res["knn purity"].append(get_knn_purity(latents, labels.reshape((-1, 1))))
    ebm = entropy_batch_mixing(latents, batch_indices)
    res["entropy batch mixing"].append(ebm[1] if isinstance(ebm, tuple) else ebm)

    cfg.input_dim = latents.shape[1]
    cfg.count_classes = np.unique(dataset.batch_indices).shape[0]
    cfg.count_labels = np.unique(dataset.labels).shape[0]

    (
        latents_train,
        latents_test,
        batches_train,
        batches_test,
        labels_train,
        labels_test,
    ) = train_test_split(
        latents,
        batch_indices,
        labels,
        test_size=0.25,
        stratify=batch_indices.reshape(-1),
    )

    latents_train = torch.Tensor(latents_train).cuda()
    latents_test = torch.Tensor(latents_test).cuda()

    batches_train_tensor = torch.zeros(latents_train.shape[0], cfg.count_classes)
    batches_train_tensor = batches_train_tensor.scatter(
        1, LongTensor(batches_train.astype("int16")).view(-1, 1), 1
    )
    batches_train_tensor = batches_train_tensor.cuda()

    labels_train_tensor = torch.zeros(latents_train.shape[0], cfg.count_labels)
    labels_train_tensor = labels_train_tensor.scatter(
        1, LongTensor(labels_train.astype("int16")).view(-1, 1), 1
    )
    labels_train_tensor = labels_train_tensor.cuda()

    train_dataset = torch.utils.data.TensorDataset(
        latents_train, batches_train_tensor, labels_train_tensor
    )
    dataloader = torch.utils.data.DataLoader(train_dataset, batch_size=cfg.batch_size)

    cfg.classifier_input_dim = cfg.bottleneck
    ohe_classifier, form_classifier = train_classifiers(
        cfg, dataloader, cfg.count_labels, cfg.count_classes
    )
    preds_batches = ohe_classifier(latents_test)
    preds_labels = form_classifier(latents_test)

    res["batch classifing accuracy"] = (
        preds_batches.argmax(1).cpu().detach().numpy() == batches_test
    ).mean()
    res["labels classifing accuracy"] = (
        preds_labels.argmax(1).cpu().detach().numpy() == labels_test
    ).mean()

    (Path(cfg.metrics_dir) / 'scVI').mkdir(parents=True, exist_ok=True)
    with open(os.path.join(Path(cfg.metrics_dir) / "scVI", log_name + ".json"), "w") as file:
        for key in res.keys():
            if type(key) is not str:
                try:
                    res[str(key)] = res[key]
                except:
                    try:
                        res[repr(key)] = res[key]
                    except:
                        raise TypeError("Unexpected key")
        json.dump(res, file)

    del vae, trainer
    del latents, batch_indices, labels, full
    del preds_batches, preds_labels, train_dataset, dataloader
    cuda.empty_cache()



from numpy import zeros, arange, ndarray
from numpy.random import randint
from torch import Tensor, randn, cuda
import pandas as pd

from pathlib import Path
from os import listdir


def create_random_classes(batch_size: int, num_classes: int):
    assert isinstance(batch_size, int), TypeError("batch size must be numeric, integer")
    assert batch_size > 0, ValueError("It can't be negative")

    assert isinstance(num_classes, int), TypeError("num_classes muust be numeric, integer")
    assert num_classes > 0, ValueError("It can't be negative")

    res_np = zeros((batch_size, num_classes))
    res_np[arange(batch_size), randint(0, num_classes, batch_size)] = 1.

    return Tensor(res_np)


def add_noise(expr, beta: float = 1.0, use_cuda: bool = False):
    #TODO doc-string
    rand_values = randn(expr.shape)
    rand_values = rand_values.to(expr.device)
    #if use_cuda and cuda.is_available():
    #    rand_values = rand_values.cuda()
    res = expr + rand_values * beta

    return res


def get_raw_data(directory: Path, preprocessing=None, indices=False):
    '''preprocess - callable'''
    assert callable(preprocessing) or preprocessing is None
    original_expr_df = pd.read_csv(directory.joinpath('ST1 - original_expression.csv'), index_col=0)
    transfer_annot_df = pd.read_csv(directory.joinpath('ST2 - transfer_annotation.csv'), index_col=0)
    transfer_annot_df = transfer_annot_df.drop_duplicates(subset='sample_ID')

    expression = original_expr_df.to_numpy()
    class_ohe = pd.get_dummies(transfer_annot_df['Init state']).to_numpy()
    cell_type = pd.get_dummies(transfer_annot_df['Cell type']).to_numpy()
    cell_indices = list(original_expr_df.index)

    if preprocessing:
        expression = preprocessing(expression)
        class_ohe = preprocessing(class_ohe)
        cell_type = preprocessing(cell_type)

    if indices:
        return expression, class_ohe, cell_type, cell_indices
    return expression, class_ohe, cell_type



import os
import stvae

# If we're using gpu accelerator
os.environ["CUDA_DEVICE_ORDER"]="PCI_BUS_ID"
os.environ["CUDA_VISIBLE_DEVICES"]="0"

dataset = stvae.datasets.MouseDataset(download=True)
# to make metrics better let's filter genes by variance
N_TOP_GENES = 700
dataset = stvae.preprocessing.get_high_variance_genes(dataset, n_genes=N_TOP_GENES)

# split data
train, test, classif = dataset.split(0.15, True, 0.15)

import torch

# the main way to set option: Use stvae.Config dataclass
cfg = stvae.Config()
cfg.count_labels = dataset.n_labels
cfg.count_classes = dataset.n_batches
cfg.use_cuda = False#torch.cuda.is_available()
cfg.input_dim = N_TOP_GENES
cfg.epochs=1
cfg.classifier_epochs=1
model = stvae.stVAE(cfg)

cfg.verbose = 'none' # output
model.train(train, validation_data=None)
metrics = model.test(test, classif)
print(metrics)



'''
Module with functions to work with scanpy AnnData objects
'''

from typing import Optional
from scanpy import AnnData
from pandas import DataFrame
import numpy as np



def _make_anndata(X: np.ndarray,
                  observation: DataFrame,
                  variables: Optional[DataFrame] = None) -> AnnData:
    '''Make a scanpy AnnData object out of pieces

        :Param X: numpy array with biological data, e.g. expression
        :Param observation: annotation for biological data
        :Param variables: some data along second dimension of expression, e.g. genes
        :Return: AnnData object
    '''
    return AnnData(X, observation, variables)


def make_anndata(expression: np.ndarray,
                 batch_index: Optional[np.ndarray] = None,
                 batch_name: Optional[str] = None,
                 cell_info: Optional[np.ndarray] = None,
                 cell_info_name: Optional[str] = None,
                 variables_info: Optional[tuple] = None) -> AnnData:
    '''Convert data to scanpy AnnData format

        :Param expression: gene expression or same thing, X at the dataset,
            shape [count_observation X cont genes]
        :Param batch_index: indeces of batches for each observation, shape
            [count_observation]
        :Param cell_info any biological information what must be saved after
            autoencoder, shape [cout_observation]
        :Param batch_name: name of the column with batch_index
        :Param cell_info_name: name of the column with cell_info

        NOTE: batch_index and cell_info will be flatten

        TODO: custom annotation data
    '''

    annotation = {}
    if not (batch_index is None and batch_name is None):
        if not (batch_index is None or batch_name is None):
            batch_index = batch_index.reshape(-1) #make observation flatten
            if batch_index.shape[0] != expression.shape[0]:
                raise ValueError(f"Count observation in expression and \
                                 batch_index must be same, got \
                                 {expression.shape[0]} != \
                                 {batch_index.shape[0]}")
            annotation[batch_name] = batch_index
        else:
            raise ValueError(f"Expected both batch_index and batch_name or \
                            noone, but got {type(batch_index)} and \
                             {type(batch_name)}")
    if not (cell_info is None and cell_info_name is None):
        if not (cell_info is None or cell_info_name is None):
            cell_info = cell_info.reshape(-1) #make it flatten for table
            if cell_info.shape[0] != expression.shape[0]:
                raise ValueError(f"Count observation in expression and \
                                    batch_index must be same, got \
                                    {expression.shape[0]} != \
                                    {batch_index.shape[0]}")

            annotation[cell_info_name] = cell_info
        else:
            raise ValueError(f"Expected both cell_info and cell_info_name or \
                            noone, but got {type(cell_info)} and \
                             {type(batch_name)}")

    observation = DataFrame(annotation)
    variables = None
    if not variables_info is None:
        variables = {}
        for name, data in variables_info:
            assert data.shape[0] == expression.shape[1], ValueError("along second dimension of expression")
            variables[name] = data
        variables = DataFrame(variables)

    return _make_anndata(expression, observation, variables)



from pathlib import Path
from ignite.engine import Engine, Events
from ignite.metrics import Metric
from ignite.contrib.handlers import ProgressBar
from numpy import inf
from tensorboardX import SummaryWriter
import torch

from .utils import save_weights, load_datasets
from .experiment import Experiment
from .config import Config
from .update import VAEUpdater, Validator
from .model.autoencoders import VAE
from .model.modules import Latent_discriminator


THIS_DIR = Path('./').absolute()
EXPERIMENTS_DIR = THIS_DIR.joinpath('experiment')


def init_weights(m):
    if type(m) == torch.nn.Linear:
        torch.nn.init.xavier_uniform(m.weight, 1e-2)
        m.bias.data.fill_(0.01)


def create_update_class(model, discriminator, config):
    is_cuda = torch.cuda.is_available() and config.use_cuda
    device = None

    if is_cuda:
        device = torch.device('cuda')
    else:
        device = torch.device('cpu')

    training_upd = VAEUpdater(model, discriminator, config, device=device,
                              cuda=is_cuda)
    validator_upd = Validator(model, discriminator, config.vae_beta,
                              device=device,
                              cuda=is_cuda)

    return training_upd, validator_upd


def create_model(cfg):
    vae_model = VAE(cfg.bottleneck, cfg.input_dim, cfg.count_classes, cfg.n_layers, cfg.scale_alpha)
    init_weights(vae_model)
    disc_model = Latent_discriminator(cfg.bottleneck, cfg.count_classes)
    init_weights(disc_model)

    if torch.cuda.is_available():
        vae_model = vae_model.cuda()
        disc_model = disc_model.cuda()

    return vae_model, disc_model


def log_progress(epoch, iteration, losses, mode='train', tensorboard_writer=None, use_iteration=False):
    if not use_iteration:
        losses_str = [
            f'{name}: {val:.3f}'
            for name, val in losses.items()
        ]
        losses_str = ' | '.join(losses_str)

        epoch_str = f'Epoch [{epoch}|{iteration}] {mode}'

        print(f'{epoch_str:<25}{losses_str}')

    for name, val in losses.items():
        tensorboard_writer.add_scalar(f'{name}', val, epoch if not use_iteration else iteration)


def train(dataloader_train, dataloader_val, cfg,
          model=None, disc=None,
          updaters=None, random_seed=None):
    with Experiment(EXPERIMENTS_DIR, cfg, random_seed=random_seed) as exp:
        print(f'Experiment started: {exp.experiment_id}')

        if not model and not disc:
            model, disc = create_model(cfg)

        if not updaters:
            update_class_train, update_class_val = create_update_class(model, disc, cfg)
        elif isinstance(updaters, (tuple, list)):
            update_class_train = updaters[0]
            update_class_val = updaters[1]
        elif isinstance(updaters, dict):
            update_class_train = updaters['train']
            update_class_val = updaters['validation']
        else:
            update_class_train = updaters
            update_class_val = Validator(model, disc, cfg, cuda=cfg.use_cuda)


        validator = Engine(update_class_val)
        trainer = Engine(update_class_train)
        progress_bar = ProgressBar()
        progress_bar.attach(trainer)

        trainer.iteration = 0

        best_loss = inf
        last_cyclic_loss = inf
        count_increas_losses = 0

        if cfg.verbose != 'silent':
            @trainer.on(Events.EPOCH_COMPLETED)
            def log_training_results(engine):
                nonlocal best_loss
                nonlocal last_cyclic_loss
                nonlocal count_increas_losses

                #trainer._process_function.model_scheduler.step()
                #trainer._process_function.latent_discrim_scheduler.step()

                if dataloader_val is None:
                    return
                validator_state = validator.run(dataloader_val)

                log_progress(trainer.state.epoch, trainer.state.iteration, validator.state.losses, 'val', tensorboard_writer)

                try:
                    #early stopping
                    if last_cyclic_loss - validator_state.losses['cyclic'] < 0.0018:
                        count_increas_losses += 1
                    else:
                        count_increas_losses = 0
                    last_cyclic_loss = validator_state.losses['cyclic']

                    if count_increas_losses >= cfg.early_stop_epochs:
                        print("=========Early stopping: loss doesn't decreases",
                            f"{cfg.early_stop_epochs} epochs=========")
                        #engine.terminate()


                    if validator_state.losses['cyclic'] < best_loss:
                        best_loss = validator_state.losses['cyclic']
                        try:
                            pass
                            #save_weights(model, exp.experiment_dir.joinpath('best_vae.pth'))
                            #save_weights(disc, exp.experiment_dir.joinpath('best_disc.pth'))
                        except Exception as e:
                            print(e)
                except:
                    # if no cyclic loss or smth
                    pass

        _TENSORBOARD_DIR = exp.experiment_dir.joinpath('log')
        tensorboard_writer = SummaryWriter(str(_TENSORBOARD_DIR))

        trainer.run(dataloader_train, max_epochs=cfg.epochs)

        print(f'Experiment {exp.experiment_id} has finished')

    return model, disc


if __name__ == '__main__':
    cfg = Config()
    dataloader_train, dataloader_val = load_datasets(cfg)
    train(dataloader_train, dataloader_val, cfg)



import scipy
from sklearn.neighbors import KNeighborsClassifier as KNN
from sklearn.neighbors import NearestNeighbors
from numpy.random import randint, choice
from numpy import stack, zeros, unique, arange, log, vectorize, logical_or
from numpy import mean as np_mean
from math import log2
from warnings import warn
from torch import Tensor
import pdb

def count2d(arr, max_ind):
    freq = zeros((arr.shape[0], max_ind+1))
    for idx, line in enumerate(arr):
        for item in line:
            freq[idx][int(item)] += 1 / (arr.shape[1] + 1e-13)

    return freq



def get_knn_purity(latents, labels, n_neighs=30):
    latents = latents.cpu().detach().numpy() if isinstance(latents, Tensor) else latents
    labels = labels.cpu().detach().numpy() if isinstance(labels, Tensor) else labels

    nbrs = NearestNeighbors(n_neighbors=n_neighs + 1).fit(latents)
    indices = nbrs.kneighbors(latents, return_distance=False)[:, 1:]
    neigh_labels = vectorize(lambda x: labels[x])(indices)

    scores = ((neigh_labels - labels.reshape(-1, 1)) == 0).mean(axis=1)
    res = [np_mean(scores[labels.reshape(scores.shape) == i]) for i in unique(labels)]

    return np_mean(res)


def get_batch_entropy(data, max_ind, classifier=None, batch_size=50):
    #TODO assert classfier.n_neighbor == batch_size
    if classifier is None:
        classifier = KNN(batch_size)

    X, Y = data[0], data[1]
    classifier.fit(X, Y)

    indeces = randint(X.shape[0], size=(batch_size,))
    points = X[indeces]
    neigh = classifier.kneighbors(points, batch_size, False)#no distances
    #neigh's shape is (num_points x batch_size)
    freqs = count2d(Y[neigh], max_ind)

    entropy_score = 0.
    for line in freqs:
        entropy_score += sum([-i*log2(i) for i in line if i != 0.])

    return entropy_score / neigh.shape[0]


def entropy_batch_mixing(latents, batches, n_neighs=50, n_pools=50, n_samples=100):
    unique_batch_ind = unique(batches)
    count_batches = unique_batch_ind.shape[0]
    if count_batches < 2:
        warn('Not enough unique batches')
        return -1.
    elif count_batches == 2:
        return entropy_batch_mixing_(latents, batches, n_neighs, n_pools, n_samples)
    else:
        batches = batches.reshape((-1,))
        average_ebm = 0.
        counter = 0
        ebm_dict = {}
        for i in unique_batch_ind:
            ebm_dict[i] = {}
            for j in unique_batch_ind:
                if i == j:
                    continue

                ebm_dict[i][j] = entropy_batch_mixing_(latents[logical_or(batches == i, batches == j)],
                                             batches[logical_or(batches == i, batches == j)])
                average_ebm += ebm_dict[i][j]
                counter += 1

        return ebm_dict, average_ebm / counter


def entropy_batch_mixing_(latents, batches, n_neighs=50, n_pools=50, n_samples=100):
    def cross_entropy(data):
        n_batches = len(unique(data))
        assert n_batches == 2, ValueError("Entropy can be calculated with only 2 batches")

        freq = np_mean(data == unique(data)[0])
        if freq == 0 or freq == 1:
            return 0

        return -freq * log(freq) - (1 - freq) * log(1 - freq)

    n_neighs = min(n_neighs, latents.shape[0] - 1)
    knn = NearestNeighbors(n_neighbors=n_neighs + 1, n_jobs=8)
    knn.fit(latents)
    kmatrix = knn.kneighbors_graph(latents) - scipy.sparse.identity(latents.shape[0])

    score = 0
    #pdb.set_trace()
    for t in range(n_pools):
        indices = choice(arange(latents.shape[0]), size=n_samples)
        while unique(batches[kmatrix[indices].nonzero()[1][kmatrix[indices].nonzero()[0] == 1]]).shape[0] < 2:
            indices = choice(arange(latents.shape[0]), size=n_samples)
        score += np_mean([
            cross_entropy(
                batches[kmatrix[indices].nonzero()[1][kmatrix[indices].nonzero()[0] == 1]]
            )
        for i in range(n_samples)])

    return score / n_pools



from typing import Sized, Callable, Union, Optional
from warnings import warn
import numpy as np
from numpy import array as Array
from torch import Tensor
from torch.utils.data import TensorDataset
from torch.utils.data import DataLoader as TorchDataLoader
from scanpy import AnnData

from .model import VAE, Latent_discriminator
from .config import Config
from .datasets import CsvDataset
from .train_script import train
from .test_script import test, train_classifiers


def _cast_data(data, cfg: Config, mode: str, datatype: str = "", preproc: bool = True):
    assert mode in ("train", "test")

    kwargs = {"batch_size": cfg.batch_size}  # for torch DataLoader
    if cfg.num_workers:
        kwargs["num_workers"] = cfg.num_workers

    if isinstance(data, (CsvDataset)):  # TODO scvi.dataset.GeneExpression
        expression = data.X
        if preproc:
            expression = np.log(expression + 1.)
        tensors = (Tensor(expression), Tensor(data.batch_indices))
        if mode == "test":
            tensors += (Tensor(data.labels),)
        ds = TensorDataset(*tensors)
        return TorchDataLoader(ds, **kwargs)
    elif isinstance(data, (tuple, list, Array)):
        items_shape = data[0].shape
        for item in data:
            if item.shape[0] != items_shape[0]:
                raise ValueError(f"Size mismatch {items_shape[0]} and {item.shape[0]}")

        '''
        if (len(data) != 2 and mode == "train") or (len(data) != 3 and mode == "test"):
            raise ValueError(
                f"Expected {datatype} data \
                with dim=2, got dim={len(data)}"
            )
        '''

        n = 3 if mode == "test" else 2
        tensors = [Tensor(data[i]) for i in range(n)]
        if preproc:
            tensors[0] = np.log(tensors[0] + 1.)
        ds = TensorDataset(*tensors)

        return TorchDataLoader(ds, **kwargs)
    elif isinstance(data, dict):
        if "expression" in data.keys():
            expression = data["expression"]
        elif "X" in data.keys():
            expression = data["X"]
        else:
            raise KeyError(f"{datatype} data must contains 'X' or 'expression' key")

        if preproc:
            expression = np.log(expression + 1.)

        try:
            batches = data["batch_indices"]
            labels = data["labels"]
        except KeyError as err:
            raise KeyError(f"{datatype} data must contains {str(err)} key")

        tensors = (expression, batches, labels)
        ds = TensorDataset(*tensors)
        return TorchDataLoader(
            ds, batch_size=cfg.batch_size, num_workers=cfg.num_workers
        )
    elif isinstance(data, AnnData):
        raise NotImplementedError()
    elif isinstance(data, TorchDataLoader):
        return data
    elif isinstance(data, TensorDataset):
        return TorchDataLoader(
            ds, batch_size=cfg.batch_size, num_workers=cfg.num_workers
        )
    else:
        raise NotImplementedError()


class stVAE:
    def __init__(self, cfg: Config):
        self.model = VAE(
            bottleneck=cfg.bottleneck,
            input_size=cfg.input_dim,
            count_classes=cfg.count_classes,
            n_layers=cfg.n_layers,
            scale_alpha=cfg.scale_alpha,
            ohe_latent_dim=cfg.condition_latent_dim,
        )
        self.disc = Latent_discriminator(cfg.bottleneck, cfg.count_classes)
        self.cfg = cfg
        self.celltype_clf = None
        self.form_clf = None
        if cfg.use_cuda:
            self.model = self.model.cuda()
            self.disc = self.disc.cuda()

    def train(
        self,
        train_data: Union[Sized, AnnData],
        validation_data: Optional[Union[Sized, AnnData]],
        random_seed=0,
        preprocess: bool = True
    ) -> None:
        train_data = _cast_data(train_data, self.cfg, "train", preproc=preprocess)
        valid_data = None
        if not validation_data is None:
            valid_data = _cast_data(validation_data, self.cfg, "train", "validation")
        train(train_data, valid_data, self.cfg, self.model, self.disc)

    def train_classifiers(self, train_data) -> None:
        train_data = _cast_data(train_data, self.cfg, "train")
        self.celltype_clf, self.form_clf = train_classifiers(
            self.cfg, train_data, self.cfg.count_labels, self.cfg.count_classes
        )

    def test(
        self,
        test_data,
        classifier_train_data=None,
        custom_metrics=[],
        overwrite: Union[bool] = False,
        preprocess: bool = True
    ) -> dict:
        if not classifier_train_data is None:
            classifier_train_data = _cast_data(classifier_train_data, self.cfg, "test", preprocess)
        test_data = _cast_data(test_data, self.cfg, "test")
        expression = test_data.dataset.tensors[0].numpy()
        batch_indices = test_data.dataset.tensors[1].numpy()
        labels = test_data.dataset.tensors[2].numpy()
        if self.celltype_clf is None or self.form_clf is None:
            if classifier_train_data is None:
                raise ValueError(
                    "If classifiers are not trained, \
                    you must provide data for training."
                )

            return test(
                self.cfg,
                self.model,
                self.disc,
                classifier_train_data,
                expression,
                batch_indices,
                labels,
            )
        else:
            if not classifier_train_data is None:
                warn(
                    'Data is specified, but classifiers already exist. \
                    New classifiers will be trained, but they will not be saved. \
                        To save new classifiers, set the flag "overwrite=True'
                )

                return test(
                    self.cfg,
                    self.model,
                    self.disc,
                    classifier_train_data,
                    expression,
                    batch_indices,
                    labels,
                )
            else:
                return test(
                    self.cfg,
                    self.model,
                    self.disc,
                    None,
                    expression,
                    batch_indices,
                    labels,
                )

    def set_config(self, cfg: Config) -> None:
        self.cfg = cfg

    def __repr__(self) -> str:
        repr = (
            "stVAE class contains\n"
            + "VAE:\n"
            + self.model.__repr__()
            + "\nDiscriminator:\n"
            + self.disc.__repr__()
            + "\nAnd configuration file "
        )
        if "name" in self.cfg.__dict__:
            repr += self.cfg.name
        return repr



#!/usr/bin/env python
# encoding: utf-8
import torch

min_var_est = 1e-8


# Consider linear time MMD with a linear kernel:
# K(f(x), f(y)) = f(x)^Tf(y)
# h(z_i, z_j) = k(x_i, x_j) + k(y_i, y_j) - k(x_i, y_j) - k(x_j, y_i)
#             = [f(x_i) - f(y_i)]^T[f(x_j) - f(y_j)]
#
# f_of_X: batch_size * k
# f_of_Y: batch_size * k
def linear_mmd2(f_of_X, f_of_Y):
    loss = 0.0
    delta = f_of_X - f_of_Y
    loss = torch.mean((delta[:-1] * delta[1:]).sum(1))
    return loss


# Consider linear time MMD with a polynomial kernel:
# K(f(x), f(y)) = (alpha*f(x)^Tf(y) + c)^d
# f_of_X: batch_size * k
# f_of_Y: batch_size * k
def poly_mmd2(f_of_X, f_of_Y, d=2, alpha=1.0, c=2.0):
    K_XX = (alpha * (f_of_X[:-1] * f_of_X[1:]).sum(1) + c)
    K_XX_mean = torch.mean(K_XX.pow(d))

    K_YY = (alpha * (f_of_Y[:-1] * f_of_Y[1:]).sum(1) + c)
    K_YY_mean = torch.mean(K_YY.pow(d))

    K_XY = (alpha * (f_of_X[:-1] * f_of_Y[1:]).sum(1) + c)
    K_XY_mean = torch.mean(K_XY.pow(d))

    K_YX = (alpha * (f_of_Y[:-1] * f_of_X[1:]).sum(1) + c)
    K_YX_mean = torch.mean(K_YX.pow(d))

    return K_XX_mean + K_YY_mean - K_XY_mean - K_YX_mean


def _mix_rbf_kernel(X, Y, sigma_list):
    assert(X.size(0) == Y.size(0))
    m = X.size(0)

    Z = torch.cat((X, Y), 0)
    ZZT = torch.mm(Z, Z.t())
    diag_ZZT = torch.diag(ZZT).unsqueeze(1)
    Z_norm_sqr = diag_ZZT.expand_as(ZZT)
    exponent = Z_norm_sqr - 2 * ZZT + Z_norm_sqr.t()

    K = 0.0
    for sigma in sigma_list:
        gamma = 1.0 / (2 * sigma**2)
        K += torch.exp(-gamma * exponent)

    return K[:m, :m], K[:m, m:], K[m:, m:], len(sigma_list)


def mix_rbf_mmd2(X, Y, sigma_list, biased=True):
    K_XX, K_XY, K_YY, d = _mix_rbf_kernel(X, Y, sigma_list)
    # return _mmd2(K_XX, K_XY, K_YY, const_diagonal=d, biased=biased)
    return _mmd2(K_XX, K_XY, K_YY, const_diagonal=False, biased=biased)


def mix_rbf_mmd2_and_ratio(X, Y, sigma_list, biased=True):
    K_XX, K_XY, K_YY, d = _mix_rbf_kernel(X, Y, sigma_list)
    # return _mmd2_and_ratio(K_XX, K_XY, K_YY, const_diagonal=d, biased=biased)
    return _mmd2_and_ratio(K_XX, K_XY, K_YY, const_diagonal=False, biased=biased)


################################################################################
# Helper functions to compute variances based on kernel matrices
################################################################################


def _mmd2(K_XX, K_XY, K_YY, const_diagonal=False, biased=False):
    m = K_XX.size(0)    # assume X, Y are same shape

    # Get the various sums of kernels that we'll use
    # Kts drop the diagonal, but we don't need to compute them explicitly
    if const_diagonal is not False:
        diag_X = diag_Y = const_diagonal
        sum_diag_X = sum_diag_Y = m * const_diagonal
    else:
        diag_X = torch.diag(K_XX)                       # (m,)
        diag_Y = torch.diag(K_YY)                       # (m,)
        sum_diag_X = torch.sum(diag_X)
        sum_diag_Y = torch.sum(diag_Y)

    Kt_XX_sums = K_XX.sum(dim=1) - diag_X             # \tilde{K}_XX * e = K_XX * e - diag_X
    Kt_YY_sums = K_YY.sum(dim=1) - diag_Y             # \tilde{K}_YY * e = K_YY * e - diag_Y
    K_XY_sums_0 = K_XY.sum(dim=0)                     # K_{XY}^T * e

    Kt_XX_sum = Kt_XX_sums.sum()                       # e^T * \tilde{K}_XX * e
    Kt_YY_sum = Kt_YY_sums.sum()                       # e^T * \tilde{K}_YY * e
    K_XY_sum = K_XY_sums_0.sum()                       # e^T * K_{XY} * e

    if biased:
        mmd2 = ((Kt_XX_sum + sum_diag_X) / (m * m)
            + (Kt_YY_sum + sum_diag_Y) / (m * m)
            - 2.0 * K_XY_sum / (m * m))
    else:
        mmd2 = (Kt_XX_sum / (m * (m - 1))
            + Kt_YY_sum / (m * (m - 1))
            - 2.0 * K_XY_sum / (m * m))

    return mmd2


def _mmd2_and_ratio(K_XX, K_XY, K_YY, const_diagonal=False, biased=False):
    mmd2, var_est = _mmd2_and_variance(K_XX, K_XY, K_YY, const_diagonal=const_diagonal, biased=biased)
    loss = mmd2 / torch.sqrt(torch.clamp(var_est, min=min_var_est))
    return loss, mmd2, var_est


def _mmd2_and_variance(K_XX, K_XY, K_YY, const_diagonal=False, biased=False):
    m = K_XX.size(0)    # assume X, Y are same shape

    # Get the various sums of kernels that we'll use
    # Kts drop the diagonal, but we don't need to compute them explicitly
    if const_diagonal is not False:
        diag_X = diag_Y = const_diagonal
        sum_diag_X = sum_diag_Y = m * const_diagonal
        sum_diag2_X = sum_diag2_Y = m * const_diagonal**2
    else:
        diag_X = torch.diag(K_XX)                       # (m,)
        diag_Y = torch.diag(K_YY)                       # (m,)
        sum_diag_X = torch.sum(diag_X)
        sum_diag_Y = torch.sum(diag_Y)
        sum_diag2_X = diag_X.dot(diag_X)
        sum_diag2_Y = diag_Y.dot(diag_Y)

    Kt_XX_sums = K_XX.sum(dim=1) - diag_X             # \tilde{K}_XX * e = K_XX * e - diag_X
    Kt_YY_sums = K_YY.sum(dim=1) - diag_Y             # \tilde{K}_YY * e = K_YY * e - diag_Y
    K_XY_sums_0 = K_XY.sum(dim=0)                     # K_{XY}^T * e
    K_XY_sums_1 = K_XY.sum(dim=1)                     # K_{XY} * e

    Kt_XX_sum = Kt_XX_sums.sum()                       # e^T * \tilde{K}_XX * e
    Kt_YY_sum = Kt_YY_sums.sum()                       # e^T * \tilde{K}_YY * e
    K_XY_sum = K_XY_sums_0.sum()                       # e^T * K_{XY} * e

    Kt_XX_2_sum = (K_XX ** 2).sum() - sum_diag2_X      # \| \tilde{K}_XX \|_F^2
    Kt_YY_2_sum = (K_YY ** 2).sum() - sum_diag2_Y      # \| \tilde{K}_YY \|_F^2
    K_XY_2_sum  = (K_XY ** 2).sum()                    # \| K_{XY} \|_F^2

    if biased:
        mmd2 = ((Kt_XX_sum + sum_diag_X) / (m * m)
            + (Kt_YY_sum + sum_diag_Y) / (m * m)
            - 2.0 * K_XY_sum / (m * m))
    else:
        mmd2 = (Kt_XX_sum / (m * (m - 1))
            + Kt_YY_sum / (m * (m - 1))
            - 2.0 * K_XY_sum / (m * m))

    var_est = (
        2.0 / (m**2 * (m - 1.0)**2) * (2 * Kt_XX_sums.dot(Kt_XX_sums) - Kt_XX_2_sum + 2 * Kt_YY_sums.dot(Kt_YY_sums) - Kt_YY_2_sum)
        - (4.0*m - 6.0) / (m**3 * (m - 1.0)**3) * (Kt_XX_sum**2 + Kt_YY_sum**2)
        + 4.0*(m - 2.0) / (m**3 * (m - 1.0)**2) * (K_XY_sums_1.dot(K_XY_sums_1) + K_XY_sums_0.dot(K_XY_sums_0))
        - 4.0*(m - 3.0) / (m**3 * (m - 1.0)**2) * (K_XY_2_sum) - (8 * m - 12) / (m**5 * (m - 1)) * K_XY_sum**2
        + 8.0 / (m**3 * (m - 1.0)) * (
            1.0 / m * (Kt_XX_sum + Kt_YY_sum) * K_XY_sum
            - Kt_XX_sums.dot(K_XY_sums_1)
            - Kt_YY_sums.dot(K_XY_sums_0))
        )
    return mmd2, var_est


from torch.nn import Module, Sequential, Linear, LeakyReLU
from torch.nn import BatchNorm1d
from torch import exp, randn_like, cat

from .modules import MishLayer

import numpy as np


def get_hidden_list(n, h, n_layers, alpha):
    return [
        int(np.round(n - (n - h) * (x / n_layers) ** alpha))
        for x in np.arange(n_layers)
    ]


class VAE(Module):
    """
    Variational autoencoder

    :param bottleneck: int, dimension of latents
    :param input_size: int, dimension of input feature vector
    :param n_layers: int, number of encoder layers
    :param scale_alpha: float, scaling factor
    :param count_classes: int
    :ohe_latent_dim: int, dimension of provided class embedding
    """

    def __init__(
        self,
        bottleneck: int,
        input_size: int,
        count_classes: int,
        n_layers: int,
        scale_alpha: float,
        ohe_latent_dim: int = 10,
    ):
        assert bottleneck > 0
        assert input_size > 0
        assert count_classes > 0
        assert ohe_latent_dim > 0

        self.bottleneck = bottleneck
        self.latent_dim = bottleneck
        self.input_size = input_size
        self.count_classes = count_classes
        self.encoder_sizes = get_hidden_list(
            input_size, bottleneck, n_layers, scale_alpha
        )

        self.decoder_sizes = [bottleneck + ohe_latent_dim] + get_hidden_list(
            input_size, bottleneck + ohe_latent_dim, n_layers, scale_alpha
        )[::-1]

        super(VAE, self).__init__()
        self.ohe_latent_size = ohe_latent_dim
        encoder_module_list = []
        for insize_, outsize_ in zip(self.encoder_sizes, self.encoder_sizes[1:]):
            encoder_module_list.extend(
                [
                    Linear(insize_, outsize_, bias=False),
                    BatchNorm1d(outsize_),
                    MishLayer(),  # LeakyReLU(0.2),
                ]
            )

        self.pre_encoder = Sequential(*encoder_module_list)

        self.fc_mu = Linear(self.encoder_sizes[-1], bottleneck, bias=False)
        self.fc_sigma = Linear(self.encoder_sizes[-1], bottleneck, bias=False)
        self.fc_class = Linear(self.encoder_sizes[-1], self.count_classes, bias=False)

        decoder_module_list = []
        for insize_, outsize_ in zip(self.decoder_sizes, self.decoder_sizes[1:]):
            decoder_module_list.extend(
                [
                    MishLayer(),
                    Linear(insize_, outsize_, bias=False),
                    BatchNorm1d(outsize_),
                ]
            )
        self.decoder = Sequential(*decoder_module_list[:-1])

        #         self.decoder = Sequential(
        #             MishLayer(),
        # #             Linear(bottleneck+self.ohe_latent_size,
        # #                    input_size
        # #                    )
        #             Linear(bottleneck+self.ohe_latent_size,
        #                     64
        #                     ),
        #             BatchNorm1d(64),
        #             MishLayer(),
        #             Linear(64,
        #                     128
        #                     ),
        #             BatchNorm1d(128),
        #             MishLayer(),
        #             Linear(128, input_size)
        #         )

        self.ohe_pipeline = Sequential(
            Linear(count_classes, self.ohe_latent_size),
            #            BatchNorm1d(self.ohe_latent_size),
            #            MishLayer(), #LeakyReLU(0.2),
            #           Linear(self.ohe_latent_size, self.ohe_latent_size),
            #           BatchNorm1d(self.ohe_latent_size),
            #           LeakyReLU(0.2),
            #           Linear(self.ohe_latent_size, self.ohe_latent_size),
            #           BatchNorm1d(self.ohe_latent_size),
            #           LeakyReLU(0.2),
            #           Linear(self.ohe_latent_size, self.ohe_latent_size),
            #           BatchNorm1d(self.ohe_latent_size),
            #           LeakyReLU(0.2),
        )

    def encode(self, x, y):
        h = self.pre_encoder(x)
        return self.fc_mu(h), self.fc_sigma(h), self.fc_class(h)

    def reparameterize(self, mu, logvar):
        std = exp(0.5 * logvar)
        eps = randn_like(std)
        return mu + eps * std

    def decode(self, z, y):
        """
            :return res, y: return decoded vectors and activations after mmd layer
        """
        y = self.ohe_pipeline(y)
        w = cat([z, y], dim=1)
        res = self.decoder(w)
        return res, y

    def forward(self, x, y):
        mu, logvar, form = self.encode(x, y)
        z = self.reparameterize(mu, logvar)
        return self.decode(z, y), mu, logvar

    def latents(self, x, y):
        mu, logvar, form = self.encode(x, y)
        return self.reparameterize(mu, logvar)



from typing import Optional, Union, Dict
from pathlib import Path
import sys
from ignite.engine import Engine, Events
from ignite.metrics import Metric
from itertools import combinations
import numpy as np
from tensorboardX import SummaryWriter
from torch.cuda import is_available as cuda_is_available
from torch import nn
from torch import Tensor
import torch
from sklearn.model_selection import train_test_split
import argparse
from tqdm import tqdm
from sklearn.metrics import confusion_matrix
import pandas as pd

from .data import get_raw_data
from .model.autoencoders import VAE
from .model.modules import Latent_discriminator
from .model import Classifier
from .config import Config
from .utils import load_datasets
from .model.classifier import Classifier
from .model.ml import get_knn_purity, get_batch_entropy, entropy_batch_mixing


THIS_DIR = Path('./').absolute()
EXPERIMENTS_DIR = THIS_DIR.joinpath('experiment')

def train_classifiers(cfg, annot_dataloader, count_celltypes, count_classes,
                      track = False, test_data = None):
    device = torch.device('cuda' if cuda_is_available() and cfg.use_cuda else 'cpu')
    celltype_clf = Classifier(inp_size=cfg.classifier_input_dim, out_size=count_celltypes)
    form_clf = Classifier(inp_size=cfg.classifier_input_dim, out_size=count_classes)
    if cfg.use_cuda and cuda_is_available():
        celltype_clf = celltype_clf.cuda()
        form_clf = form_clf.cuda()

    celltype_clf_opt = torch.optim.Adam(celltype_clf.parameters(), 
                                        weight_decay=cfg.celltype_clf_wdecay,
                                        lr=cfg.celltype_clf_lr)
    form_clf_opt = torch.optim.Adam(form_clf.parameters(),
                                        weight_decay=cfg.form_clf_wdecay,
                                        lr=cfg.form_clf_lr)

    celltype_criterion = nn.CrossEntropyLoss()
    form_criterion = nn.CrossEntropyLoss()

    print('\n')
    metrics = {
        'celltype_losses':   [],
        'form_losses':       [],
        'celltype_accuracy': [],
        'form_accuracy':     [],
        'epoch':             []
    }
    for epoch in range(cfg.classifier_epochs):
        if cfg.verbose == 'all':
            print(f'\rTraining classifier [{epoch+1}/{cfg.classifier_epochs}]', end='')
        celltype_clf.train()
        form_clf.train()
        metrics['epoch'].append(epoch + 1)
        celltype_av_loss = 0.
        form_av_loss = 0.
        counter = 0
        for exp_, form_, cell_type_ in annot_dataloader:
            exp_ = exp_.to(device)
            form_ = form_.argmax(-1).to(device)
            cell_type_ = cell_type_.argmax(-1).to(device)

            predicted_celltype_ = celltype_clf(exp_)
            celltype_loss_on_batch = celltype_criterion(predicted_celltype_, cell_type_)
            celltype_av_loss += celltype_loss_on_batch.item()

            celltype_clf_opt.zero_grad()
            celltype_loss_on_batch.backward(retain_graph=True)
            celltype_clf_opt.step()

            predicted_form_ = form_clf(exp_)
            form_loss_on_batch = form_criterion(predicted_form_, form_)
            form_av_loss += form_loss_on_batch.item()

            form_clf_opt.zero_grad()
            form_loss_on_batch.backward(retain_graph=True)
            form_clf_opt.step()

            counter += 1

        metrics['celltype_losses'].append(celltype_av_loss / counter)
        metrics['form_losses'].append(form_av_loss / counter)
    if track:
        celltype_clf.eval()
        form_clf.eval()
        counter = 0
        celltype_accuracy = 0.
        form_accuracy = 0.
        for exp_, form_, cell_type_ in test_data:
            exp_ = exp_.to(device)
            form_ = form_.argmax(-1)
            cell_type_ = cell_type_.argmax(-1)

            predicted_celltype_ = celltype_clf(exp_).cpu().detach().numpy().argmax(-1)
            cell_type_ = cell_type_.cpu().detach().numpy()
            celltype_accuracy += (predicted_celltype_ == cell_type_).mean()

            predicted_form_ = form_clf(exp_).cpu().detach().numpy().argmax(-1)
            form_ = form_.cpu().detach().numpy()
            form_accuracy += (predicted_form_ == form_).mean()

            counter += 1
        metrics['celltype_accuracy'].append(celltype_accuracy / counter)
        metrics['form_accuracy'].append(form_accuracy / counter)

    if track:
        return celltype_clf, form_clf, metrics
    return celltype_clf, form_clf

def test(cfg, vae_model, discrim, annot_dataloader,
         test_expression,
         class_ohe_test, celltype_test,
         pretrained_classifiers=None,
         dataset_name: str = '') -> Dict[str, Union[float, str, None]]:
    '''Calculate metrics and return dict
        :Param cfg: Config dataclass
        :Param vae_model: autoecoder with declarated signature
        :Param discrim: deprecated, unused
        :Param annot_dataloader: dataloader(expression, batch_indices, labels)
            for classifier training
        :Param test_expression:
    '''
    #STYLE TRANSFER
    ge_transfer_raw = np.repeat(test_expression, class_ohe_test.shape[1], axis=0)
    ge_transfer_raw = Tensor(ge_transfer_raw)
    init_classes_transfer = np.repeat(class_ohe_test, class_ohe_test.shape[1], axis=0)
    init_classes_transfer = Tensor(init_classes_transfer)
    init_celltypes_transfer = np.repeat(celltype_test, class_ohe_test.shape[1], axis=0)
    init_celltypes_transfer = Tensor(init_celltypes_transfer)

    device = torch.device('cpu')
    if cfg.use_cuda and cuda_is_available():
        device = torch.device('cuda')
        ge_transfer_raw = ge_transfer_raw.cuda()
        init_classes_transfer = init_classes_transfer.cuda()
        init_celltypes_transfer = init_celltypes_transfer.cuda()

    target_classes_transfer = np.zeros((class_ohe_test.shape[0] * class_ohe_test.shape[1],
                                        class_ohe_test.shape[1]))
    target_classes_transfer[np.arange(target_classes_transfer.shape[0]),
                        np.arange(target_classes_transfer.shape[0])%target_classes_transfer.shape[1]] = 1

    target_classes_transfer = Tensor(target_classes_transfer)
    if cfg.use_cuda and cuda_is_available():
        target_classes_transfer = target_classes_transfer.cuda()

    transfer_expression_tensor = vae_model(ge_transfer_raw, target_classes_transfer)[0]
    if isinstance(transfer_expression_tensor, tuple):
        transfer_expression_tensor = transfer_expression_tensor[0]
    transfer_expression_np = transfer_expression_tensor.cpu().detach().numpy()

    test_expression_tensor = Tensor(test_expression)
    class_ohe_test_tensor = Tensor(class_ohe_test)
    if cfg.use_cuda and cuda_is_available():
        target_classes_transfer = target_classes_transfer.cuda()
        test_expression_tensor = test_expression_tensor.cuda()
        class_ohe_test_tensor = class_ohe_test_tensor.cuda()
    reconstruction = vae_model(test_expression_tensor, class_ohe_test_tensor)[0]
    if isinstance(reconstruction, tuple):
        reconstruction = reconstruction[0]
    reconstruction = reconstruction.cpu().detach().numpy()

    mse = (reconstruction - test_expression) ** 2

    residual_transfer = ge_transfer_raw.cpu().numpy() - transfer_expression_np
    res_norm = (residual_transfer**2).mean(1).reshape((-1, cfg.count_classes))#l2 norm,
    res_equal = class_ohe_test.argmax(1)
    res_equal = res_equal if isinstance(res_equal, np.ndarray) else res_equal.cpu().detach().numpy()

    if cfg.verbose in ('all', 'test'):
        print('Mean square error reconstructed expression:', np.mean(mse))
        print('Mean values of test expression:', test_expression.mean())
        print('Mean values of reconstructed expression:', reconstruction.mean())

        print('part of test expression <0.5 mean:', (test_expression < 0.5).mean())
        print('part reconstructed exrression <0.5 mean:', (reconstruction < 0.5).mean())
        print('Calibration accuracy', (res_norm.argmin(1) == res_equal).mean())
        print('\n')


    test_latents = vae_model.latents(test_expression_tensor, class_ohe_test_tensor)
    transfered_latents = vae_model.latents(ge_transfer_raw, target_classes_transfer)

    #train
    celltype_clf_expr = None
    celltype_clf_latents = None
    form_clf_expr = None
    form_clf_latents = None
    if not isinstance(test_latents, torch.Tensor):
        test_latents = torch.Tensor(test_latents)
    if not isinstance(transfered_latents, torch.Tensor):
        transfered_latents = torch.Tensor(transfered_latents)
    if cfg.use_cuda and cuda_is_available():
        test_latents = test_latents.cuda()
        transfered_latents = transfered_latents.cuda()

    if pretrained_classifiers is None:
        cfg.classifier_input_dim = cfg.input_dim
        celltype_clf_expr, form_clf_expr = train_classifiers(
                                                    cfg, annot_dataloader,
                                                    celltype_test.shape[1],
                                                    cfg.count_classes
                                                )
        cfg.classifier_input_dim = vae_model.latent_dim
        latents_dataset = torch.utils.data.TensorDataset(
            test_latents,
            class_ohe_test_tensor,
            torch.randn(test_latents.shape[0], cfg.count_labels)
        )
        latents_dataloader = torch.utils.data.DataLoader(latents_dataset,
                                                         batch_size=cfg.batch_size,
                                                         shuffle=True,
                                                         drop_last=True)
        celltype_clf_latents, form_clf_latents = train_classifiers(
                                                    cfg, latents_dataloader,
                                                    celltype_test.shape[1],
                                                    cfg.count_classes
                                                )
    else:
        celltype_clf = pretrained_classifiers[0]
        form_clf = pretrained_classifiers[1]


    if isinstance(test_latents, Tensor):
        test_latents_np = test_latents.cpu().detach().numpy()
    if isinstance(transfered_latents, Tensor):
        transfered_latents_np = transfered_latents.cpu().detach().numpy()

    # entropy batch mixing
    ebm_score = {}
    ebm_score['test'] = [0. for i in range(1)]
    ebm_score['transfered'] = [0. for i in range(1)]
    for n in range(1):
        for lat, bind, key in zip(
                       (test_latents_np, transfered_latents_np),
                       (class_ohe_test_tensor, target_classes_transfer),
                       ebm_score.keys()):

            batch_ind = bind.argmax(1).cpu().numpy()
            ind = list(combinations(np.unique(batch_ind), 2))
            for i in ind:
                a = (batch_ind == i[0])
                b = (batch_ind == i[1])
                condition = np.logical_or(a, b)

                #Important breakpoint
                ebm_score['test'][n] += entropy_batch_mixing(
                                  lat[condition], batch_ind[condition]
                              )
                ebm_score['transfered'][n] += entropy_batch_mixing(
                                         lat[condition], batch_ind[condition]
                                       )

            if len(ind) > 0:
                ebm_score[key][n] /= len(ind)


    # KNN Purity
    test_purity = []
    transfered_purity = []
    test_purity.append(get_knn_purity(test_latents_np, celltype_test.argmax(1)))
    transfered_purity.append(get_knn_purity(transfered_latents_np, target_classes_transfer.argmax(dim=1).cpu().numpy()))


    celltype_train_labelenc = None
    if not annot_dataloader is None:
        celltype_train_labelenc = annot_dataloader.dataset.tensors[2].argmax(1).cpu().numpy()
    celltype_test_raw_labelenc = celltype_test.argmax(1)
    celltype_test_labelenc = init_celltypes_transfer.argmax(1).cpu().numpy()
    #form_test_labelenc = init_classes_transfer.argmax(1).cpu().numpy()
    form_test_labelenc = target_classes_transfer.argmax(1).cpu().numpy()

    predicted_celltype_test_raw = celltype_clf_expr(test_expression_tensor)
    predicted_celltype_test = celltype_clf_expr(transfer_expression_tensor)
    predicted_celltype_train = None
    if not annot_dataloader is None:
        annot_expression = annot_dataloader.dataset.tensors[0].to(device)
        predicted_celltype_train = celltype_clf_expr(annot_expression)
    predicted_form_test = form_clf_expr(transfer_expression_tensor)
    #predict celltypes and batch indices by model's latents
    predicted_celltype_test_latents = celltype_clf_latents(test_latents)
    predicted_celltype_transfer_latents = celltype_clf_latents(transfered_latents)
    predicted_form_latents = form_clf_latents(transfered_latents)

    # classifier on expression
    predicted_celltype_test_labelenc = predicted_celltype_test.argmax(1).cpu().detach().numpy()
    predicted_celltype_test_raw_labelenc = predicted_celltype_test_raw.argmax(1).cpu().detach().numpy()
    predicted_celltype_train_labelenc = None
    if not predicted_celltype_train is None:
        predicted_celltype_train_labelenc = predicted_celltype_train.argmax(1).cpu().detach().numpy()

    sameclass_mask = (init_classes_transfer.argmax(1) == target_classes_transfer.argmax(1)).cpu().numpy().astype('bool')

    celltype_train_accuracy = None
    if not predicted_celltype_train_labelenc is None:
        celltype_train_accuracy = (predicted_celltype_train_labelenc == celltype_train_labelenc).mean()
    celltype_test_raw_accuracy = (predicted_celltype_test_raw_labelenc == np.array(celltype_test_raw_labelenc)).mean()
    celltype_test_accuracy = (predicted_celltype_test_labelenc == celltype_test_labelenc).mean()

    celltype_notransfer_accuracy = (predicted_celltype_test_labelenc == celltype_test_labelenc).mean(0)
    recon_confusion = confusion_matrix(celltype_test_labelenc[sameclass_mask],
                                       predicted_celltype_test_labelenc[sameclass_mask])
    if cfg.verbose in ('all', 'test'):
        print('Confusion matrix | CELLTYPE RECONSTRUCTION')
        print(recon_confusion)

    from sklearn.metrics import classification_report
    rep = classification_report(
        celltype_test_labelenc[sameclass_mask],
        predicted_celltype_test_labelenc[sameclass_mask],
        output_dict=True
    )
    df = pd.DataFrame(rep).transpose()
    if cfg.verbose in ('all', 'test'):
        print(df)
        print('Celltype reconstruction report')

    if (~sameclass_mask).any():
        celltype_transfer_accuracy = (predicted_celltype_test_labelenc[~sameclass_mask] == celltype_test_labelenc[~sameclass_mask]).mean(0)
        transfer_confusion = confusion_matrix(celltype_test_labelenc[~sameclass_mask],
                                              predicted_celltype_test_labelenc[~sameclass_mask])
        if cfg.verbose in ('all', 'test'):
            print('Confusion matrix | CELLTYPE TRANSFER')
            print(transfer_confusion)

        rep = classification_report(
            celltype_test_labelenc[~sameclass_mask],
            predicted_celltype_test_labelenc[~sameclass_mask],
            output_dict=True
        )
        df = pd.DataFrame(rep).transpose()
        if cfg.verbose in ('all', 'test'):
            print('Celltype transfer report')
            print(df)

        pred_f_t = predicted_form_test if isinstance(predicted_form_test, np.ndarray) else predicted_form_test.argmax(1).cpu().detach().numpy()
        rep = classification_report(
            form_test_labelenc[~sameclass_mask],
            pred_f_t[~sameclass_mask],
            output_dict=True
        )
        df = pd.DataFrame(rep).transpose()
        if cfg.verbose in ('all', 'test'):
            print('Form transfer report')
            print(df)
    else:
        celltype_transfer_accuracy = "Not enoug classes for transfering"
    predicted_form_test = predicted_form_test.argmax(1).cpu().detach().numpy()
    form_test_accuracy = (predicted_form_test == form_test_labelenc).mean(0)
    recon_confusion = confusion_matrix(form_test_labelenc[sameclass_mask],
                                       predicted_form_test[sameclass_mask],
                                       normalize='all')
    if cfg.verbose in ('all', 'test'):
        print('Confusion matrix | FORM RECONSTRUCTED')
        print(recon_confusion)

    transfer_confusion = confusion_matrix(form_test_labelenc[~sameclass_mask],
                                          pred_f_t[~sameclass_mask],
                                          normalize='all')
    if cfg.verbose in ('all', 'test'):
        print('Confusion matrix | FORM TRANSFER')
        print(transfer_confusion)

    rep = classification_report(
        form_test_labelenc[~sameclass_mask],
        pred_f_t[~sameclass_mask],
        output_dict=True
    )
    df = pd.DataFrame(rep).transpose()
    if cfg.verbose in ('all', 'test'):
        print('Form reconstructed report')
        print(df)


    # classifier on latents
    predicted_celltype_latents_labelenc = predicted_celltype_test_latents.argmax(1).cpu().detach().numpy()
    predicted_celltype_transfer_latents_labelenc = predicted_celltype_transfer_latents.argmax(1).cpu().detach().numpy()

    #celltype_test_latents_accuracy = (predicted_celltype_latents_labelenc == celltype_test_raw_labelenc).mean()
    celltype_notransfer_latents_accuracy = (np.array(predicted_celltype_latents_labelenc) == np.array(celltype_test_raw_labelenc)).mean(0)
    if (~sameclass_mask).any():
        celltype_transfer_latents_accuracy = (predicted_celltype_transfer_latents_labelenc[~sameclass_mask] == celltype_test_labelenc[~sameclass_mask]).mean()
    else:
        celltype_transfer_latents_accuracy = "Not enoug classes for transfering"

    #reconstructed means transfered to the same class

    if cfg.verbose in ('all', 'test'):
        print('On expression:')
        print('\tCell type prediction accuracy [vae transfered data]', celltype_test_accuracy,
                '< with the same class transfer')
        print('\tCell type prediction accuracy [reconstructed data]', celltype_notransfer_accuracy)
        print('\tCell type prediction accuracy [transfered to another classes]', celltype_transfer_accuracy)
        print('\tClass prediction accuracy:', form_test_accuracy)
        print('On latents:')
        #print('\tCell type prediction accuracy [vae transfered data]', celltype_test_latents_accuracy,
        #        '< with the same class transfer')
        print('\tCell type prediction accuracy [reconstructed data]', celltype_notransfer_latents_accuracy)
        print('\tCell type prediction accuracy [transfered to another classes]', celltype_transfer_latents_accuracy)

        #print('K neighbors purity [vae test data]:', test_purity)
        #print('K neighbors purity [transfer]:', transfered_purity)
        #print('Entropy batch mixing [vae test data]:', ebm_score['test'])
        #print('Entropy batch mixing [transfered data]:', ebm_score['transfered'])


    return {
        'MSE': float(np.mean(mse)),
        'Calibration accuracy': float((res_norm.argmin(1) == res_equal).mean()),
        'Cell type prediction accuracy [train]:': celltype_train_accuracy,
        'Cell type prediction accuracy expression [test hold out data]': celltype_test_raw_accuracy,
        #'Cell type prediction accuracy latents [test hold out data]': celltype_test_latents_accuracy,
        #'Cell type prediction accuracy expression [vae transfered data]': celltype_test_accuracy,
        #'Cell type prediction accuracy latents [vae transfered data]': celltype_test_accuracy,
        'Cell type prediction accuracy expression [reconstructed data]': celltype_notransfer_accuracy,
        'Cell type prediction accuracy latents [reconstructed data]': celltype_notransfer_latents_accuracy,
        'Cell type prediction accuracy expression [transfered to another classes]': celltype_transfer_accuracy,
        'Cell type prediction accuracy latents [transfered to another classes]': celltype_transfer_latents_accuracy,
        'Class prediction accuracy': form_test_accuracy,
        #'K neighbors purity [vae test data]': test_purity,
        #'K neighbors purity [transfer]': transfered_purity,
        #'Entropy batch mixing [test data]': ebm_score['test'],
        #'Entropy batch mixing [transfer data]': ebm_score['transfered']
    }


if __name__ == '__main__':
    pass



import os
import requests
from pathlib import Path
from typing import Optional, Union, List
import ntpath
import pandas as pd
import numpy as np
from sklearn.preprocessing import OneHotEncoder
from sklearn.model_selection import train_test_split


def filename_from_path(path):
    head, tail = ntpath.split(path)
    return tail or ntpath.basename(head)


class CsvDataset:
    def __init__(self, names: List[Union[Path, str]]):
        dfs = [pd.read_csv(filename) for filename in names]
        self.assign(*dfs)
        self.filter(*dfs)

    def assign(self, expression_df, batches_df, labels_df):
        # self.labels = LabelEncoder().fit_transform(labels_df.to_numpy())
        labels_np = labels_df.to_numpy()
        self.labels = OneHotEncoder(sparse=False).fit_transform(
            labels_np.reshape(-1, 1)
        )
        self.n_labels = np.unique(labels_np).shape[0]
        # self.batch_indices = LabelEncoder().fit_transform(batches_df.to_numpy())
        batch_np = batches_df.to_numpy()
        self.batch_indices = OneHotEncoder(sparse=False).fit_transform(
            batch_np.reshape(-1, 1)
        )
        self.n_batches = np.unique(batch_np).shape[0]
        self.X = expression_df.to_numpy()
        self.nb_genes = self.X.shape[1]

    def filter(self, expression_df, batches_df, labels_df):
        for checking in (self.labels, self.batch_indices):
            labels, counts = np.unique(checking, return_counts=True)
            for idx, count in enumerate(counts):
                if count < 2:
                    cell_name = labels_df[labels_df.label == labels[idx]].index[0]
                    expression_df = expression_df[expression_df.index != cell_name]
                    labels_df = labels_df[labels_df.index != cell_name]
                    batches_df = batches_df[batches_df.index != cell_name]
        self.assign(expression_df, batches_df, labels_df)

    def __len__(self):
        return self.X.shape[0]

    def split(
        self,
        test_size: float = 0.15,
        for_classifiers: bool = True,
        classif_size: Optional[float] = None,
    ) -> tuple:
        """Split data into train and test sets.
        It optional splits train data into vae train data and classifirs train data
        :param test_size: float part of test size
        :param for_classifiers: bool if True returns data for classifiers training additionally"""
        (
            train_expression,
            test_expression,
            train_batches,
            test_batches,
            train_labels,
            test_labels,
        ) = train_test_split(
            self.X, self.batch_indices, self.labels, test_size=test_size,
            stratify=self.batch_indices
        )

        if for_classifiers:
            (
                train_expression,
                class_expression,
                train_batches,
                class_batches,
                train_labels,
                class_labels,
            ) = train_test_split(
                train_expression, train_batches, train_labels, test_size=classif_size,
                stratify=train_batches
            )
            return (
                (train_expression, train_batches, train_labels),
                (test_expression, test_batches, test_labels),
                (class_expression, class_batches, class_labels)
            )
        else:
            return (
                (train_expression, train_batches, train_labels),
                (test_expression, test_batches, test_labels)
            )


class ScratchDataset(CsvDataset):
    def __init__(
        self, expression: np.ndarray, batch_indices: np.ndarray, labels: np.ndarray
    ):
        self.X = expression
        self.nb_genes = self.X.shape[1]
        self.labels = labels
        if len(labels.shape) == 1:
            self.n_labels = np.unique(labels).shape[0]
        elif len(labels.shape) == 2:
            self.n_labels = labels.shape[1]
        self.batch_indices = batch_indices
        if len(batch_indices.shape) == 1:
            self.n_batches = np.unique(batch_indices).shape[0]
        elif len(batch_indices.shape) == 2:
            self.n_batches = batch_indices.shape[1]

    def __len__(self):
        return self.X.shape[0]


class MouseDataset(CsvDataset):
    """To work with mouse single-cell expression dataset.
    Contains the ways to read/download data and process it

    To download data provide download flag as True
    If you already have data provide a full path for each one

    :param directory: str or pathlib.Path full path to the directory
                        contains data files. If it's None but filenames provided,
                        class tries to find at './' catalog
    :param expression_filename:
    :param batch_filename:
    :param labels_filename:
    :param download: bool, if download flag is True class downloads dataset files.
        If directory is None it downloads to './'
    """

    def __init__(
        self,
        directory: Optional[Union[Path, str]] = None,
        expression_filename: Optional[Union[Path, str]] = None,
        batch_filename: Optional[Union[Path, str]] = None,
        labels_filename: Optional[Union[Path, str]] = None,
        download: bool = False,
        verbose: bool = True,
    ):
        self.cell_attribute_names = {
            "labels",
            "local_vars",
            "local_means",
            "batch_indices",
        }
        if not (
            directory
            or expression_filename
            or batch_filename
            or labels_filename
            or download
        ):
            if not (
                os.path.isfile("./expression.csv")
                and os.path.isfile("./batches.csv")
                and os.path.isfile("./labels.csv")
            ):
                raise ValueError(
                    "There is not enough data to find the dataset. \
                    To download data, set the flag download = True."
                )

        directory = Path(directory) if directory else Path("./")

        if expression_filename is None:
            expression_filename = directory / "expression.csv"
        else:
            if not isinstance(expression_filename, Path):
                expression_filename = Path(expression_filename)
            if not directory in expression_filename.parents:
                expression_filename = directory / expression_filename

        if isinstance(directory, str):
            directory = Path(directory)
        if batch_filename is None:
            batch_filename = directory / "batches.csv"
        else:
            if not isinstance(batch_filename, Path):
                batch_filename = Path(batch_filename)
            if not directory in batch_filename.parents:
                batch_filename = directory / batch_filename

        if labels_filename is None:
            labels_filename = directory / "labels.csv"
        else:
            if not isinstance(labels_filename, Path):
                labels_filename = Path(labels_filename)
            if not directory in labels_filename.parents:
                labels_filename = directory / labels_filename

        if download:
            self.download(
                "https://ndownloader.figshare.com/files/22334787",
                expression_filename,
                verbose,
            )
            self.download(
                "https://ndownloader.figshare.com/files/22332978",
                batch_filename,
                verbose,
            )
            self.download(
                "https://ndownloader.figshare.com/files/22332975",
                labels_filename,
                verbose,
            )

        expression_df = pd.read_csv(expression_filename, index_col=0)

        labels_df = pd.read_csv(labels_filename, index_col=0)
        assert all(expression_df.index == labels_df.index)

        batches_df = pd.read_csv(batch_filename, index_col=0)
        assert all(expression_df.index == batches_df.index)

        dfs = (expression_df, batches_df, labels_df)
        self.assign(*dfs)
        self.filter(*dfs)

    def download(self, url: str, filename: Union[Path, str], verbose: bool) -> None:
        if verbose:
            print(f"Downloading {filename_from_path(filename)}")
        response = requests.get(url, allow_redirects=True, stream=verbose)
        length = response.headers.get("content-length")
        with open(filename, "wb") as file:
            if not length and verbose:
                file.write(response.content)
            else:
                dl = 0
                length = int(length)
                from tqdm import tqdm
                from math import ceil

                for data in tqdm(
                    response.iter_content(chunk_size=4096),
                    total=ceil(length / 4096),
                    position=0,
                    leave=True,
                ):
                    file.write(data)
                    dl += len(data)



from data import make_anndata
from utils import load_datasets, scvi_anndata, get_high_variance_genes
from train_script import train
from test_script import test
from config import Config
from model.trvae import trVAE
from experiment import Experiment

from scvi.dataset import (
                    PreFrontalCortexStarmapDataset,
                    FrontalCortexDropseqDataset,
                    RetinaDataset,
                    HematoDataset,
                    AnnDatasetFromAnnData
                  )
import os
import json
import numpy as np
from torch import zeros, Tensor, LongTensor, log, cuda, save, load
import scanpy as sc
import trvaep
from functools import partial
from pathlib import Path


def benchmark_trvae(dataset, log_name, cfg, **kwargs):
    ds = dataset
    n_genes = min(ds.X.shape[1], cfg.n_genes)

    scvai_genes, scvai_batches_ind, scvai_labels_ind = get_high_variance_genes(
        ds.X,
        ds.batch_indices,
        ds.labels,
        n_genes = n_genes,
        argmax=False
    )
    cfg.count_classes = int(np.max(ds.batch_indices) + 1)
    cfg.count_labels = int(np.max(ds.labels) + 1)
    cfg.input_dim = int(scvai_genes.shape[1])


    data = load_datasets(cfg, True, True,
                         (scvai_genes, scvai_batches_ind, scvai_labels_ind),
                         0.9)
    dataloader_train = data[0]
    dataloader_val = data[1]
    dataloader_test = data[2]
    annot_train = data[3]
    annot_test = data[4]
    x, batch_ind, celltype = annot_train.dataset.tensors
    batch_ind = batch_ind.argmax(dim=1)
    celltype = celltype.argmax(dim=1)

    anndata_train = make_anndata(x.cpu().numpy(),
                                 batch_ind.cpu().numpy(),
                                 'condition',
                                 celltype.cpu().numpy(),
                                 'cell_type')
    x_test, batch_ind_test, celltype_test = annot_test.dataset.tensors
    batch_ind_test = batch_ind_test.argmax(dim=1)
    celltype_test = celltype_test.argmax(dim=1)
    anndata_test = make_anndata(x_test.cpu().numpy(), batch_ind_test.cpu().numpy(),
                                'condition', celltype_test.cpu().numpy(), 'cell_type')
    sc.pp.normalize_per_cell(anndata_train)
    sc.pp.normalize_per_cell(anndata_test)
    sc.pp.log1p(anndata_train)
    sc.pp.log1p(anndata_test)

    n_conditions = anndata_train.obs["condition"].unique().shape[0]
    x_test = anndata_test.X
    batch_ind_test_tmp = anndata_test.obs['condition']
    batch_ind_test = zeros(batch_ind_test_tmp.shape[0], cfg.count_classes)
    batch_ind_test = batch_ind_test.scatter(1, LongTensor(batch_ind_test_tmp.astype('uint16')).view(-1, 1), 1).numpy()
    celltype_test_tmp = anndata_test.obs['cell_type']
    celltype_test = zeros(celltype_test_tmp.shape[0], cfg.count_labels)
    celltype_test = celltype_test.scatter(1, LongTensor(celltype_test_tmp.astype('uint16')).view(-1, 1), 1).numpy()

    model = trVAE(x.shape[1],
                  num_classes=n_conditions,
                  encoder_layer_sizes=[128, 32],
                  decoder_layer_sizes=[32, 128],
                  latent_dim=cfg.bottleneck,
                  alpha=0.0001,
                 )
    trainer = trvaep.Trainer(model, anndata_train)

    print('Training...')
    trainer.train_trvae(cfg.epochs, 512, 50)#n_epochs, batch_size, early_patience

    print('Tests...')
    print('Dataset:', log_name)
    res = test(cfg,
                model, None,
                annot_train,
                x_test,
                batch_ind_test,
                celltype_test
    )
    res['n_genes'] = n_genes

    metrics_path = Path(cfg.metrics_dir) / 'trVAE'
    metrics_path.mkdir(parents=True, exist_ok=True)
    with open(metrics_path / (log_name + '.json'), 'w') as file:
        json.dump(res, file, indent=4)

    del ds
    del model
    del data
    del dataloader_train, dataloader_val, dataloader_test
    del annot_train, annot_test
    del scvai_genes, scvai_batches_ind, scvai_labels_ind
    cuda.empty_cache()




from dataclasses import dataclass
from pathlib import Path
from numpy import inf
from typing import Optional, Tuple, Union


@dataclass
class Config:
    '''
    Training configuration
    '''
    noise_beta: float = 0.1
    decay_beta: float = 1.
    clip_value: float = 1.
    vae_lr: float = 0.001
    vae_decay: float = 1e-3
    disc_lr: float = 1e-4
    disc_decay: float = 0.
    vae_beta: float = 2e-5
    adv_weight: float = 0.0000001
    cyclic_weight: float = 0.2
    mmd_weight: float = 0
    l1_weight: float = 0#.5
    kernel_mu: float = 0.4
    model_scheduler_gamma: float = 0.992
    discr_scheduler_gamma: float = 0.992
    n_layers: int = 2
    scale_alpha: float = 1.3
    form_consistency_weight: float = 0.2

    bottleneck: int = 30
    input_dim: int = 15987
    count_labels: int = 12
    count_classes: int = 3
    epochs: int = 600
    batch_size: int = 128
    num_workers: Optional[int] = None
    activation: str = 'mish'
    n_genes: int = 700
    condition_latent_dim: int = 10

    classifier_hidden_size: int = 512
    classifier_epochs: int = 450
    celltype_clf_lr: float = 1e-5
    form_clf_lr: float = 3e-4
    celltype_clf_wdecay: float = 0#weight decay
    form_clf_wdecay: float = 0#weight decay

    experiment_dir: Union[Path, str] = Path('./data/experiment')
    data_dir: Union[Path, str] = Path('./genes')
    metrics_dir: Union[Path, str] = Path('./metrics')
    best_loss = inf
    early_stop_epochs: int = 30#Number of epochs to wait if no loss improvement and then stop the training.
    use_cuda: bool = True
    random_state: int = 512
    cuda: bool = True
    reproducibility: bool = True
    verbose: str = 'all'
    device_ids: Tuple[int] = (2,)

    def __str__(self):
        return '\n'.join([f"{key}: {value}" for key, value in self.__dict__.items()])



from utils import load_datasets, get_high_variance_genes, scvi_anndata
from train_script import train
from test_script import test
from config import Config
from experiment import Experiment
from model.autoencoders import VAE
from data import get_raw_data

from scvi.dataset import (
                    PreFrontalCortexStarmapDataset,
                    RetinaDataset,
                  )
import argparse
import os
import json
import numpy as np
from torch import zeros, Tensor, LongTensor, log, cuda, save, load
import scanpy as sc
from functools import partial
from pathlib import Path
import gc


def benchmark_stvae(dataset, log_name, cfg, **kwargs):
    ds = dataset
    n_genes = min(ds.X.shape[1], cfg.n_genes)
    expression = np.log(ds.X + 1.)
    scvai_genes, scvai_batches_ind, scvai_labels_ind = get_high_variance_genes(
        expression,
        ds.batch_indices,
        ds.labels,
        n_genes=n_genes,
        argmax=False
    )

    cfg.count_classes = np.unique(ds.batch_indices).shape[0]
    cfg.count_labels = np.unique(ds.labels).shape[0]
    cfg.input_dim = int(scvai_genes.shape[1])

    data = load_datasets(cfg, True, True, (scvai_genes, scvai_batches_ind, scvai_labels_ind))
    dataloader_train = data[0]
    dataloader_val = data[1]
    dataloader_test = data[2]
    annot_train = data[3]
    annot_test = data[4]

    styletransfer_test_expr = annot_test.dataset.tensors[0].cpu().numpy()
    styletransfer_test_class = annot_test.dataset.tensors[1].cpu().numpy()
    styletransfer_test_celltype = annot_test.dataset.tensors[2].cpu().numpy()


    model = None
    disc = None

    print('Training...')
    model, disc = train(dataloader_train, dataloader_val, cfg, model, disc)

    print('Tests...')
    print('Dataset:', log_name)
    cfg.classifier_epochs=1
    res = test(cfg,
        model, disc,
        annot_train,
        styletransfer_test_expr,
        styletransfer_test_class,
        styletransfer_test_celltype,
        dataset_name=log_name
    )

    (Path(cfg.metrics_dir) / 'stVAE').mkdir(parents=True, exist_ok=True)
    with open(Path(cfg.metrics_dir) / 'stVAE/' / (log_name + '.json'), 'w') as file:
        json.dump(res, file, indent=4)

    del ds
    del model, disc
    del styletransfer_test_expr
    del styletransfer_test_class
    del styletransfer_test_celltype
    del data
    del dataloader_train, dataloader_val, dataloader_test
    del annot_train, annot_test
    del scvai_genes, scvai_batches_ind, scvai_labels_ind
    gc.collect()
    cuda.empty_cache()



import os
import argparse
import numpy as np
import torch
from torch import nn, optim

import torch.nn.functional as F
from torch.autograd import Variable
from torch.utils.data import DataLoader

import torchvision
from torchvision.utils import save_image
from torchvision import datasets, transforms


def get_hidden_list(n, h, n_layers, alpha):
    return [int(np.round(n - (n-h)*(x/n_layers)**alpha)) for x in np.arange(n_layers)]


class DCDiscriminator(nn.Module):
    def __init__(self, inp_size: int, image_bottleneck: int,
                    n_img_layers: int, labels_dim: int,
                    alpha: float = 1.):
        assert isinstance(inp_size, int) and int > 0

        super(ModelD, self).__init__()
        x_layers_list = []
        for insize_, outsize_ in zip(self.encoder_sizes, self.encoder_sizes[1:]):
            x_layers_list.extend([
                Linear(insize_, outsize_),
                MishLayer(),
                BatchNorm1d(outsize_),
            ])

        self.x_pipeline = nn.Sequential(*x_layers_list)
        self.conditional_pipeline = nn.Sequential(
            nn.Linear(image_bottleneck + labels_dim, (image_bottleneck + labels_dim) // 2),
            MishLayer(),
            BatchNorm1d((image_bottleneck + labels_dim) // 2),
        )

        self.y_linear = nn.Sequential(
            nn.Linear(labels_dim, max(inp_size // 10, labels_dim + 1)),
            MishLayer(),
            BatchNorm1d((image_bottleneck + labels_dim) // 2),
        )
        self.fc = nn.Linear((image_bottleneck + labels_dim) // 2, 1)

    def forward(self, x, labels):
        x = self.x_pipeline(x)
        y_ = self.y_linear(labels)
        x = torch.cat([x, y_], 1)
        x = self.conditional_pipeline(x)
        x = self.fc(x)
        return F.sigmoid(x)

class ModelG(nn.Module):
    def __init__(self, z_dim):
        self.z_dim = z_dim
        super(ModelG, self).__init__()
        self.fc2 = nn.Linear(10, 1000)
        self.fc = nn.Linear(self.z_dim+1000, 64*28*28)
        self.bn1 = nn.BatchNorm2d(64)
        self.deconv1 = nn.ConvTranspose2d(64, 32, 5, 1, 2)
        self.bn2 = nn.BatchNorm2d(32)
        self.deconv2 = nn.ConvTranspose2d(32, 1, 5, 1, 2)

    def forward(self, x, labels):
        batch_size = x.size(0)
        y_ = self.fc2(labels)
        y_ = F.relu(y_)
        x = torch.cat([x, y_], 1)
        x = self.fc(x)
        x = x.view(batch_size, 64, 28, 28)
        x = self.bn1(x) 
        x = F.relu(x)
        x = self.deconv1(x)
        x = self.bn2(x)
        x = F.relu(x)
        x = self.deconv2(x)
        x = F.sigmoid(x)
        return x


from torch import(
                  mean, randn, cuda, 
                  transpose, matmul, exp
                )
from torch import sum as tsum
from torch.nn import MSELoss
from functools import partial
from .mmd import mix_rbf_mmd2


def pairwise_distance(x, y):
    if not len(x.shape) == len(y.shape) == 2:
        raise ValueError('Both inputs should be matrices.')

    if x.shape[1] != y.shape[1]:
        raise ValueError('The number of features should be the same.')

    x = x.view(x.shape[0], x.shape[1], 1)
    y = transpose(y, 0, 1)
    output = tsum((x - y) ** 2, 1)
    output = transpose(output, 0, 1)

    return output


def gaussian_kernel_matrix(x, y, gamma):
    '''
        Do the gaussian kernel-based estimation of distance
        between distributions over x and y variables
        :param sigmas: sigmas is the hyperparameter sets prior to the kernel computation
    '''
    #sigmas = sigmas.view(sigmas.shape[0], 1)
    #gamma = 1. / (2. * sigmas)
    dist = pairwise_distance(x, y).contiguous()
    dist_ = dist.view(1, -1)
    s = matmul(gamma, dist_)

    return tsum(exp(-s), 0).view_as(dist)


def maximum_mean_discrepancy(x, y, kernel):
    value = kernel(x, x) - 2 * kernel(x, y) + kernel(y, y)
    return value


def mmd_criterion(x, y, use_cuda=True, mu=1, sigma=1e-6):
    '''sigmas = randn(x.shape[0], 1)*sigma + mu
    if use_cuda and cuda.is_available:
        sigmas = sigmas.cuda()

    kernel = partial(gaussian_kernel_matrix, gamma=sigmas)
    loss = maximum_mean_discrepancy(x, y, kernel)
    '''
    loss = mix_rbf_mmd2(x, y, [1e3, 2e3, 4e3, 8e3, 16e3])

    return loss


class VariationalLoss:
    def __init__(self, vae_beta: float, rec_coef: float = 1.):
        self.reconstruction_loss = MSELoss(reduction='mean')
        self.vae_beta = vae_beta
        self.rec_coef = rec_coef

    def __call__(self, recon_x, x, mu, logvar):
        '''
        Reconstruction + KL divergence losses summed over all elements and batch.
        '''
        #BCE = F.binary_cross_entropy(recon_x, x, reduction='sum')
        rec_loss_ = self.reconstruction_loss(recon_x, x)
        # see Appendix B from VAE paper:
        # Kingma and Welling. Auto-Encoding Variational Bayes. ICLR, 2014
        # https://arxiv.org/abs/1312.6114
        # 0.5 * sum(1 + log(sigma^2) - mu^2 - sigma^2)
        KLD = -0.5 * mean(1 + logvar - mu.pow(2) - logvar.exp())

        return self.rec_coef*rec_loss_ + self.vae_beta*KLD




from collections import Iterable, Sized
from typing import Union, Optional
import logging
import pickle
import torch
import numpy as np
from numpy import max
import scanpy as sc
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder

from .data import get_raw_data, make_anndata
from .datasets import ScratchDataset


def save_pickle(obj, filename):
    with open(filename, "wb") as f:
        pickle.dump(obj, f)


def load_pickle(filename):
    with open(filename, "rb") as f:
        obj = pickle.load(f)

    return obj


def load_weights(model, filename):
    # load trained on GPU models to CPU
    if not torch.cuda.is_available():

        def map_location(storage, loc):
            return storage

    else:
        map_location = None

    state_dict = torch.load(str(filename), map_location=map_location)

    if isinstance(model, torch.nn.DataParallel):
        model = model.module

    model.load_state_dict(state_dict)


def save_weights(model, filename):
    if isinstance(model, torch.nn.DataParallel):
        model = model.module

    torch.save(model.state_dict(), str(filename))


def init_weights(modules):
    if isinstance(modules, torch.nn.Module):
        modules = modules.modules()

    for m in modules:
        if isinstance(m, torch.nn.Sequential):
            init_weights(m_inner for m_inner in m)

        if isinstance(m, torch.nn.ModuleList):
            init_weights(m_inner for m_inner in m)

        if isinstance(m, torch.nn.Linear):
            m.reset_parameters()
            torch.nn.init.xavier_normal_(m.weight.data)
            # m.bias.data.zero_()
            if m.bias is not None:
                m.bias.data.normal_(0, 0.01)

        if isinstance(m, torch.nn.Conv2d):
            torch.nn.init.xavier_normal_(m.weight.data)
            m.bias.data.zero_()

        if isinstance(m, torch.nn.Conv1d):
            torch.nn.init.xavier_normal_(m.weight.data)
            m.bias.data.zero_()


def to_device(obj, device=None):
    if device is None:
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    if isinstance(obj, (list, tuple)):
        return [to_device(o, device) for o in obj]

    if isinstance(obj, dict):
        return {k: to_device(o, device) for k, o in obj.items()}

    if isinstance(obj, np.ndarray):
        obj = torch.from_numpy(obj)

    obj = obj.to(device)
    return obj


def load_datasets(
    cfg,
    test: bool = False,
    annot: bool = False,
    custom_data=None,
    preprocessing=None,
    test_size: float = 0.25,
    indices=None,
):
    expr, class_ohe, cell_type = None, None, None
    if custom_data is None:
        expr, class_ohe, cell_type = get_raw_data(cfg.data_dir, preprocessing)
    else:
        expr, class_ohe, cell_type = custom_data

    assert not (
        expr is None or class_ohe is None or cell_type is None
    ), "Any data have not to be None"
    if indices is None:
        indices = np.array(expr.shape[0] * [0])

    (
        train_expression,
        val_expression,
        train_class_ohe,
        val_class_ohe,
        train_annot,
        test_annot,
        train_ind,
        test_ind,
    ) = train_test_split(
        expr,
        class_ohe,
        cell_type,
        indices,
        random_state=cfg.random_state,
        stratify=class_ohe.argmax(1),
        test_size=test_size,
    )

    full_test_expr = torch.Tensor(val_expression)
    full_test_form = torch.Tensor(val_class_ohe)

    if test:
        (
            val_expression,
            test_expression,
            val_class_ohe,
            test_class_ohe,
            val_ind,
            test_ind,
        ) = train_test_split(
            val_expression,
            val_class_ohe,
            test_ind,
            random_state=cfg.random_state,
            test_size=0.15,
        )
        test_expression_tensor = torch.Tensor(test_expression)
        test_class_ohe_tensor = torch.Tensor(test_class_ohe)

    val_expression_tensor = torch.Tensor(val_expression)
    val_class_ohe_tensor = torch.Tensor(val_class_ohe)
    train_expression_tensor = torch.Tensor(train_expression)
    train_class_ohe_tensor = torch.Tensor(train_class_ohe)
    train_annot_tensor = torch.Tensor(train_annot)
    test_annot_tensor = torch.Tensor(test_annot)

    if cfg.cuda and torch.cuda.is_available():
        val_expression_tensor = val_expression_tensor.cuda()
        val_class_ohe_tensor = val_class_ohe_tensor.cuda()
        train_expression_tensor = train_expression_tensor.cuda()
        train_class_ohe_tensor = train_class_ohe_tensor.cuda()
        if test:
            test_expression_tensor = test_expression_tensor.cuda()
            test_class_ohe_tensor = test_class_ohe_tensor.cuda()
        full_test_expr = full_test_expr.cuda()
        train_annot_tensor = train_annot_tensor.cuda()
        test_annot_tensor = test_annot_tensor.cuda()
        full_test_form = full_test_form.cuda()

    trainset = torch.utils.data.TensorDataset(
        train_expression_tensor, train_class_ohe_tensor
    )
    dataloader_train = torch.utils.data.DataLoader(
        trainset,
        batch_size=cfg.batch_size,
        shuffle=True,
        # num_workers=cfg.num_workers,
        drop_last=True,
    )
    valset = torch.utils.data.TensorDataset(val_expression_tensor, val_class_ohe_tensor)
    dataloader_val = torch.utils.data.DataLoader(
        valset, batch_size=val_expression_tensor.size(0), shuffle=False, drop_last=True
    )

    result = tuple((dataloader_train, dataloader_val))

    if test:
        testset = torch.utils.data.TensorDataset(
            test_expression_tensor, test_class_ohe_tensor
        )
        dataloader_test = torch.utils.data.DataLoader(
            testset,
            batch_size=test_expression_tensor.size(0),
            shuffle=False,
            drop_last=True,
        )

        result += tuple((dataloader_test,))

    if annot:
        annot_dataset_train = torch.utils.data.TensorDataset(
            train_expression_tensor, train_class_ohe_tensor, train_annot_tensor
        )
        annot_dataloader_train = torch.utils.data.DataLoader(
            annot_dataset_train, batch_size=cfg.batch_size, shuffle=True, drop_last=True
        )

        annot_dataset_test = torch.utils.data.TensorDataset(
            full_test_expr, full_test_form, test_annot_tensor
        )
        annot_dataloader_test = torch.utils.data.DataLoader(
            annot_dataset_test,
            batch_size=full_test_expr.size(0),
            shuffle=False,
            drop_last=True,
        )

        result += tuple((annot_dataloader_train, annot_dataloader_test))

    if not indices is None:
        result += tuple((train_ind, test_ind))

    return result


def get_high_variance_genes(
    expression: np.ndarray,
    class_ohe: np.ndarray,
    cell_type: np.ndarray,
    genes: Optional[Union[list, tuple, np.ndarray]] = None,
    n_genes: Optional[int] = None,
    min_disp: Optional[float] = None,
    max_disp: Optional[float] = None,
    argmax: bool = True,
) -> tuple:
    assert n_genes or min_disp or max_disp, "Choose smth"

    genes_data = None
    if not genes is None:
        genes_data = [("genes", genes)]
        if not isinstance(genes, np.ndarray):
            genes = np.array(genes)
    if argmax:
        class_ohe = class_ohe.argmax(1)
        cell_type = cell_type.argmax(1)
    anndata = make_anndata(
        expression, class_ohe, "condition", cell_type, "cell_type", genes_data
    )
    sc.pp.filter_genes_dispersion(
        anndata, n_top_genes=n_genes, min_disp=min_disp, max_disp=max_disp
    )

    expr = anndata.X
    expr = expr if isinstance(expr, np.ndarray) else np.array(expr)
    ohe = np.array(anndata.obs["condition"].tolist())
    ohe = OneHotEncoder(sparse=False).fit_transform(ohe.reshape(-1, 1))
    celltype = np.array(anndata.obs["cell_type"])
    celltype = OneHotEncoder(sparse=False).fit_transform(celltype.reshape(-1, 1))
    if not genes is None:
        return expr, ohe, celltype, anndata.var["genes"]

    return expr, ohe, celltype


def scvi_anndata(dataset_class):
    """Make scanpy.Anndata from scvi dataset class"""
    ds = dataset_class
    return make_anndata(ds.X, ds.batch_indices, "condition", ds.labels, "cell_type")




import os
from collections import Iterable
import torch
import numpy as np
import tempfile
from pathlib import Path

from .utils import save_pickle, load_pickle


class Experiment:
    _CONFIG_FILENAME = 'config.pkl'

    def __init__(self, experiments_dir, config, prefix=None, random_seed=0):
        self.config = config
        self.experiments_dir = Path(experiments_dir)
        self.prefix = prefix
        self.random_seed = random_seed if isinstance(random_seed, int) else 0

        # create dir for the experiment
        if self.prefix is not None:
            self.prefix = f'{self.prefix}.'

        self.experiment_dir = None
        self.experiment_id = None

    def __enter__(self):
        # check if dir exists
        if not self.experiments_dir.is_dir():
            self.experiments_dir.mkdir(parents=True, exist_ok=True)

        self.experiment_dir = Path(tempfile.mkdtemp(dir=self.experiments_dir, prefix=self.prefix))
        self.experiment_id = self.experiment_dir.name

        # save the config file
        Experiment._save_config(self.config, self.experiment_dir)

        #make it reproducibility
        if self.config.reproducibility:
            np.random.seed(self.random_seed)
            torch.backends.cudnn.deterministic = True
            torch.backends.cudnn.benchmark = False
            torch.manual_seed(self.random_seed)

        # use cuda
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        try:
            if isinstance(self.config.device_ids, Iterable):
                device_ids = list(map(str, self.config.device_ids))
            else:
                device_ids = [str(self.config.device_ids)]
        except TypeError as err:
            raise TypeError("device_ids must contains values that can be \
                            casted into a string")
        os.environ["CUDA_VISIBLE_DEVICES"] = ','.join(device_ids)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    @classmethod
    def load(cls, experiments_dir, experiment_id):
        experiment_dir = experiments_dir.joinpath(experiment_id)

        config = Experiment._load_config(experiment_dir)

        exp = Experiment(experiments_dir, config)
        exp.experiment_dir = experiment_dir
        exp.experiment_id = exp.experiment_dir.name

        return exp

    @classmethod
    def _save_config(cls, config, experiment_dir):
        filename = experiment_dir.joinpath(Experiment._CONFIG_FILENAME)
        save_pickle(config, filename)

    @classmethod
    def _load_config(cls, experiment_dir):
        filename = experiment_dir.joinpath(Experiment._CONFIG_FILENAME)
        config = load_pickle(filename)

        return config



#!/usr/bin/env python3
from torch.nn import Module, Sequential, Linear, BatchNorm1d
from torch.nn import LeakyReLU, LogSoftmax
from torch.nn.functional import softplus
from torch import rand, add, floor, tanh
from torch import LongTensor

import math
import torch
from torch.optim.optimizer import Optimizer

class BatchSwapNoise(Module):
    '''
    Swap Noise module

    :param p: float, probability
    '''

    def __init__(self, p):
        super().__init__()
        assert p >= 0. and p <= 1., ValueError("The probability value should be in between 0 and 1.")
        self.p = p

    def forward(self, x):
        if self.training:
            mask = rand(x.size()) > (1 - self.p)
            idx = add(torch.arange(x.nelement()),
                            (floor(rand(x.size()) * x.size(0)).type(LongTensor) *
                             (mask.type(LongTensor) * x.size(1))).view(-1))
            idx[idx>=x.nelement()] = idx[idx>=x.nelement()]-x.nelement()
            return x.view(-1)[idx].view(x.size())
        else:
            return x


class Latent_discriminator(Module):
     def __init__(self, bottleneck: int, count_classes: int):
        assert bottleneck > 0, ValueError(f"A bottleneck dimension can't be negative: {bottleneck}")
        assert isinstance(bottleneck, int), TypeError("INT!")
        assert count_classes > 0, ValueError(f"A dimension can't be negative: {count_classes}")
        assert isinstance(count_classes, int), TypeError("INT!")

        super(Latent_discriminator,self).__init__()

        self.model = Sequential(
            Linear(bottleneck, 1024, bias=False),
            BatchNorm1d(1024),
            LeakyReLU(0.2, inplace=True),
            Linear(1024, 1024, bias=False),
            BatchNorm1d(1024),
            LeakyReLU(0.2, inplace=True),
            Linear(1024, count_classes, bias=False),
            LogSoftmax(dim=-1)
        )


     def forward(self,x):
        return self.model(x)



class RAdam(Optimizer):

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0):
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        self.buffer = [[None, None, None] for ind in range(10)]
        super(RAdam, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(RAdam, self).__setstate__(state)

    def step(self, closure=None):

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:

            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()
                if grad.is_sparse:
                    raise RuntimeError('RAdam does not support sparse gradients')

                p_data_fp32 = p.data.float()

                state = self.state[p]

                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p_data_fp32)
                    state['exp_avg_sq'] = torch.zeros_like(p_data_fp32)
                else:
                    state['exp_avg'] = state['exp_avg'].type_as(p_data_fp32)
                    state['exp_avg_sq'] = state['exp_avg_sq'].type_as(p_data_fp32)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']

                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                exp_avg.mul_(beta1).add_(1 - beta1, grad)

                state['step'] += 1
                buffered = self.buffer[int(state['step'] % 10)]
                if state['step'] == buffered[0]:
                    N_sma, step_size = buffered[1], buffered[2]
                else:
                    buffered[0] = state['step']
                    beta2_t = beta2 ** state['step']
                    N_sma_max = 2 / (1 - beta2) - 1
                    N_sma = N_sma_max - 2 * state['step'] * beta2_t / (1 - beta2_t)
                    buffered[1] = N_sma

                    # more conservative since it's an approximated value
                    if N_sma >= 5:
                        step_size = math.sqrt((1 - beta2_t) * (N_sma - 4) / (N_sma_max - 4) * (N_sma - 2) / N_sma * N_sma_max / (N_sma_max - 2)) / (1 - beta1 ** state['step'])
                    else:
                        step_size = 1.0 / (1 - beta1 ** state['step'])
                    buffered[2] = step_size

                if group['weight_decay'] != 0:
                    p_data_fp32.add_(-group['weight_decay'] * group['lr'], p_data_fp32)

                # more conservative since it's an approximated value
                if N_sma >= 5:            
                    denom = exp_avg_sq.sqrt().add_(group['eps'])
                    p_data_fp32.addcdiv_(-step_size * group['lr'], exp_avg, denom)
                else:
                    p_data_fp32.add_(-step_size * group['lr'], exp_avg)

                p.data.copy_(p_data_fp32)

        return loss

class PlainRAdam(Optimizer):

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0):
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)

        super(PlainRAdam, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(PlainRAdam, self).__setstate__(state)

    def step(self, closure=None):

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:

            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()
                if grad.is_sparse:
                    raise RuntimeError('RAdam does not support sparse gradients')

                p_data_fp32 = p.data.float()

                state = self.state[p]

                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p_data_fp32)
                    state['exp_avg_sq'] = torch.zeros_like(p_data_fp32)
                else:
                    state['exp_avg'] = state['exp_avg'].type_as(p_data_fp32)
                    state['exp_avg_sq'] = state['exp_avg_sq'].type_as(p_data_fp32)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']

                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                exp_avg.mul_(beta1).add_(1 - beta1, grad)

                state['step'] += 1
                beta2_t = beta2 ** state['step']
                N_sma_max = 2 / (1 - beta2) - 1
                N_sma = N_sma_max - 2 * state['step'] * beta2_t / (1 - beta2_t)

                if group['weight_decay'] != 0:
                    p_data_fp32.add_(-group['weight_decay'] * group['lr'], p_data_fp32)

                # more conservative since it's an approximated value
                if N_sma >= 5:                    
                    step_size = group['lr'] * math.sqrt((1 - beta2_t) * (N_sma - 4) / (N_sma_max - 4) * (N_sma - 2) / N_sma * N_sma_max / (N_sma_max - 2)) / (1 - beta1 ** state['step'])
                    denom = exp_avg_sq.sqrt().add_(group['eps'])
                    p_data_fp32.addcdiv_(-step_size, exp_avg, denom)
                else:
                    step_size = group['lr'] / (1 - beta1 ** state['step'])
                    p_data_fp32.add_(-step_size, exp_avg)

                p.data.copy_(p_data_fp32)

        return loss


class AdamW(Optimizer):

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0, warmup = 0):
        defaults = dict(lr=lr, betas=betas, eps=eps,
                        weight_decay=weight_decay, warmup = warmup)
        super(AdamW, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(AdamW, self).__setstate__(state)

    def step(self, closure=None):
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:

            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()
                if grad.is_sparse:
                    raise RuntimeError('Adam does not support sparse gradients, please consider SparseAdam instead')

                p_data_fp32 = p.data.float()

                state = self.state[p]

                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p_data_fp32)
                    state['exp_avg_sq'] = torch.zeros_like(p_data_fp32)
                else:
                    state['exp_avg'] = state['exp_avg'].type_as(p_data_fp32)
                    state['exp_avg_sq'] = state['exp_avg_sq'].type_as(p_data_fp32)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']

                state['step'] += 1

                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                exp_avg.mul_(beta1).add_(1 - beta1, grad)

                denom = exp_avg_sq.sqrt().add_(group['eps'])
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                
                if group['warmup'] > state['step']:
                    scheduled_lr = 1e-8 + state['step'] * group['lr'] / group['warmup']
                else:
                    scheduled_lr = group['lr']

                step_size = scheduled_lr * math.sqrt(bias_correction2) / bias_correction1
                
                if group['weight_decay'] != 0:
                    p_data_fp32.add_(-group['weight_decay'] * scheduled_lr, p_data_fp32)

                p_data_fp32.addcdiv_(-step_size, exp_avg, denom)

                p.data.copy_(p_data_fp32)

        return loss


class MishLayer(Module):
    '''
    Mish activation functio
    '''
    def __init__(self):
        super(MishLayer, self).__init__()

    def forward(self, x):
        return x*tanh(softplus(x))
